@@ -483,13 +483,16 @@ rb_playlist_add_m3u (RBPlaylist *playlist, const char *url, gpointer data)
 	g_free (contents);
 
 	for (i = 0; lines[i] != NULL; i++) {
+		if (lines[i][0] == '#' || lines[i][0] == '\0')
+			continue;
 		/* Either it's a URI, or it has a proper path ... */
 		if (strstr(lines[i], "://") != NULL
 				|| lines[i][0] == G_DIR_SEPARATOR) {
 			/* We use the same code for .ram and m3u playlists,
 			 * and .ram files can contain .smil entries */
-			if (rb_playlist_parse (playlist, lines[i]) == TRUE)
-				retval = TRUE;
+			rb_playlist_parse (playlist, lines[i]);
+			rb_playlist_add_one_url (playlist, lines[i], NULL);
+			retval = TRUE;
 		} else if (lines[i][0] == '\\' && lines[i][1] == '\\') {
 			/* ... Or it's in the windows smb form
 			 * (\\machine\share\filename), Note drive names
@@ -1039,3 +1042,27 @@ rb_playlist_parse (RBPlaylist *playlist, const char *url)
 
 	return FALSE;
 }
+
+gboolean
+rb_playlist_can_handle (RBPlaylist *playlist, const char *url)
+{
+	const char *mimetype;
+	int i;
+
+	g_return_val_if_fail (url != NULL, FALSE);
+
+	mimetype = gnome_vfs_get_mime_type (url);
+
+	if (mimetype == NULL)
+		return FALSE;
+
+	for (i = 0; i < G_N_ELEMENTS(special_types); i++)
+		if (strcmp (special_types[i].mimetype, mimetype) == 0)
+			return TRUE;
+
+	for (i = 0; i < G_N_ELEMENTS(dual_types); i++)
+		if (strcmp (dual_types[i].mimetype, mimetype) == 0)
+			return TRUE;
+
+	return FALSE;
+}
