@@ -1,7 +1,7 @@
 /*
  *  arch-tag: Implementation of RhythmDB query result GtkTreeModel
  *
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,14 +30,15 @@
 #include "rhythmdb-model.h"
 #include "rhythmdb-query-model.h"
 #include "rb-debug.h"
+#include "rb-thread-helpers.h"
 #include "gsequence.h"
-#include "eggtreemultidnd.h"
+#include "rb-tree-dnd.h"
 
 static void rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass);
 static void rhythmdb_query_model_tree_model_init (GtkTreeModelIface *iface);
 static void rhythmdb_query_model_rhythmdb_model_init (RhythmDBModelIface *iface);
-static void rhythmdb_query_model_multi_drag_source_init (EggTreeMultiDragSourceIface *iface);
-static void rhythmdb_query_model_drag_dest_init (GtkTreeDragDestIface *iface);
+static void rhythmdb_query_model_drag_source_init (RbTreeDragSourceIface *iface);
+static void rhythmdb_query_model_drag_dest_init (RbTreeDragDestIface *iface);
 static void rhythmdb_query_model_init (RhythmDBQueryModel *shell_player);
 static void rhythmdb_query_model_finalize (GObject *object);
 static void rhythmdb_query_model_set_property (GObject *object,
@@ -48,6 +49,12 @@ static void rhythmdb_query_model_get_property (GObject *object,
 					       guint prop_id,
 					       GValue *value,
 					       GParamSpec *pspec);
+static GSequencePtr rhythmdb_query_model_do_insert (RhythmDBQueryModel *model,
+						    RhythmDBEntry *entry,
+						    gboolean lock);
+static void rhythmdb_query_model_do_delete (RhythmDBQueryModel *model,
+					    RhythmDBEntry *entry,
+					    gboolean lock);
 static void rhythmdb_query_model_entry_added_cb (RhythmDB *db, RhythmDBEntry *entry,
 						 RhythmDBQueryModel *model);
 static void rhythmdb_query_model_entry_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
@@ -57,21 +64,20 @@ static void rhythmdb_query_model_entry_deleted_cb (RhythmDB *db, RhythmDBEntry *
 static gboolean rhythmdb_query_model_entry_to_iter (RhythmDBModel *rmodel, RhythmDBEntry *entry,
 						    GtkTreeIter *iter);
 static void rhythmdb_query_model_cancel (RhythmDBModel *model);
-static gboolean rhythmdb_query_model_sortable (RhythmDBModel *model);
 static gboolean rhythmdb_query_model_has_pending_changes (RhythmDBModel *model);
 static gboolean rhythmdb_query_model_poll (RhythmDBModel *model, GTimeVal *timeout);
 
-static gboolean rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource, 
-							  GList *paths, 
+static gboolean rhythmdb_query_model_drag_data_get (RbTreeDragSource *dragsource,
+							  GList *paths,
 							  GtkSelectionData *selection_data);
-static gboolean rhythmdb_query_model_multi_drag_data_delete (EggTreeMultiDragSource *dragsource,
+static gboolean rhythmdb_query_model_drag_data_delete (RbTreeDragSource *dragsource,
 							     GList *paths);
-static gboolean rhythmdb_query_model_multi_row_draggable (EggTreeMultiDragSource *dragsource,
+static gboolean rhythmdb_query_model_row_draggable (RbTreeDragSource *dragsource,
 							  GList *paths);
-static gboolean rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
+static gboolean rhythmdb_query_model_drag_data_received (RbTreeDragDest *drag_dest,
 							 GtkTreePath *dest,
 							 GtkSelectionData  *selection_data);
-static gboolean rhythmdb_query_model_row_drop_possible (GtkTreeDragDest *drag_dest,
+static gboolean rhythmdb_query_model_row_drop_possible (RbTreeDragDest *drag_dest,
 							GtkTreePath *dest,
 							GtkSelectionData  *selection_data);
 static GtkTreeModelFlags rhythmdb_query_model_get_flags (GtkTreeModel *model);
@@ -111,6 +117,7 @@ struct RhythmDBQueryModelUpdate
 		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED,
 		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED,
 		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_DELETED,
+		RHYTHMDB_QUERY_MODEL_UPDATE_QUERY_COMPLETE,
 	} type;
 	RhythmDBEntry *entry;
 };
@@ -126,13 +133,25 @@ struct RhythmDBQueryModelPrivate
 
 	guint stamp;
 
-	guint max_size;
+	guint max_size_mb;
+	guint max_count;
+
+	gboolean cancelled;
+	
+	gboolean connected;
+
+	glong total_duration;
+	GnomeVFSFileSize total_size;
 
 	GSequence *entries;
 	GHashTable *reverse_map;
-	
+
+	GAsyncQueue *query_complete;
+
 	/* row_inserted/row_changed/row_deleted */
 	GAsyncQueue *pending_updates;
+
+	gboolean reorder_drag_and_drop;
 };
 
 enum
@@ -143,6 +162,7 @@ enum
 	PROP_SORT_FUNC,
 	PROP_SORT_DATA,
 	PROP_MAX_SIZE,
+	PROP_MAX_COUNT,
 };
 
 enum
@@ -189,8 +209,8 @@ rhythmdb_query_model_get_type (void)
 			NULL
 		};
 
-		static const GInterfaceInfo multi_drag_source_info = {
-			(GInterfaceInitFunc) rhythmdb_query_model_multi_drag_source_init,
+		static const GInterfaceInfo drag_source_info = {
+			(GInterfaceInitFunc) rhythmdb_query_model_drag_source_init,
 			NULL,
 			NULL
 		};
@@ -214,11 +234,11 @@ rhythmdb_query_model_get_type (void)
 					     &rhythmdb_model_info);
 
 		g_type_add_interface_static (rhythmdb_query_model_type,
-					     EGG_TYPE_TREE_MULTI_DRAG_SOURCE,
-					     &multi_drag_source_info);
+					     RB_TYPE_TREE_DRAG_SOURCE,
+					     &drag_source_info);
 
 		g_type_add_interface_static (rhythmdb_query_model_type,
-					     GTK_TYPE_TREE_DRAG_DEST,
+					     RB_TYPE_TREE_DRAG_DEST,
 					     &drag_dest_info);
 	}
 
@@ -275,7 +295,15 @@ rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass)
 					 PROP_MAX_SIZE,
 					 g_param_spec_int ("max-size",
 							   "maxsize",
-							   "maximum size",
+							   "maximum size (MB)",
+							   0, G_MAXINT, 0,
+							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
+	g_object_class_install_property (object_class,
+					 PROP_MAX_COUNT,
+					 g_param_spec_int ("max-count",
+							   "maxcount",
+							   "maximum count (songs)",
 							   0, G_MAXINT, 0,
 							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
@@ -312,23 +340,23 @@ rhythmdb_query_model_rhythmdb_model_init (RhythmDBModelIface *iface)
 	iface->entry_to_iter = rhythmdb_query_model_entry_to_iter;
 	iface->poll = rhythmdb_query_model_poll;
 	iface->cancel = rhythmdb_query_model_cancel;
-	iface->sortable = rhythmdb_query_model_sortable;
 	iface->has_pending_changes = rhythmdb_query_model_has_pending_changes;
 }
 
 static void
-rhythmdb_query_model_multi_drag_source_init (EggTreeMultiDragSourceIface *iface)
+rhythmdb_query_model_drag_source_init (RbTreeDragSourceIface *iface)
 {
-	iface->row_draggable = rhythmdb_query_model_multi_row_draggable;
-	iface->drag_data_delete = rhythmdb_query_model_multi_drag_data_delete;
-	iface->drag_data_get = rhythmdb_query_model_multi_drag_data_get;
+	iface->row_draggable = rhythmdb_query_model_row_draggable;
+	iface->drag_data_delete = rhythmdb_query_model_drag_data_delete;
+	iface->drag_data_get = rhythmdb_query_model_drag_data_get;
 }
 
 static void
-rhythmdb_query_model_drag_dest_init (GtkTreeDragDestIface *iface)
+rhythmdb_query_model_drag_dest_init (RbTreeDragDestIface *iface)
 {
 	iface->drag_data_received = rhythmdb_query_model_drag_data_received;
 	iface->row_drop_possible = rhythmdb_query_model_row_drop_possible;
+	iface->row_drop_position = NULL;
 }
 
 static void
@@ -372,7 +400,10 @@ rhythmdb_query_model_set_property (GObject *object,
 		model->priv->sort_user_data = g_value_get_pointer (value);
 		break;
 	case PROP_MAX_SIZE:
-		model->priv->max_size = g_value_get_int (value);
+		model->priv->max_size_mb = g_value_get_int (value);
+		break;
+	case PROP_MAX_COUNT:
+		model->priv->max_count = g_value_get_int (value);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -380,7 +411,7 @@ rhythmdb_query_model_set_property (GObject *object,
 	}
 }
 
-static void 
+static void
 rhythmdb_query_model_get_property (GObject *object,
 				   guint prop_id,
 				   GValue *value,
@@ -403,7 +434,10 @@ rhythmdb_query_model_get_property (GObject *object,
 		g_value_set_pointer (value, model->priv->sort_user_data);
 		break;
 	case PROP_MAX_SIZE:
-		g_value_set_int (value, model->priv->max_size);
+		g_value_set_int (value, model->priv->max_size_mb);
+		break;
+	case PROP_MAX_COUNT:
+		g_value_set_int (value, model->priv->max_count);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -421,7 +455,10 @@ rhythmdb_query_model_init (RhythmDBQueryModel *model)
 	model->priv->entries = g_sequence_new (NULL);
 	model->priv->reverse_map = g_hash_table_new (g_direct_hash, g_direct_equal);
 
+	model->priv->query_complete = g_async_queue_new ();
 	model->priv->pending_updates = g_async_queue_new ();
+
+	model->priv->reorder_drag_and_drop = FALSE;
 }
 
 static void
@@ -451,6 +488,7 @@ rhythmdb_query_model_finalize (GObject *object)
 		}
 	}
 
+	g_async_queue_unref (model->priv->query_complete);
 	g_async_queue_unref (model->priv->pending_updates);
 
 	g_free (model->priv);
@@ -480,10 +518,36 @@ rhythmdb_query_model_new_empty (RhythmDB *db)
 			     "db", db, NULL);
 }
 
-void
-rhythmdb_query_model_complete (RhythmDBQueryModel *model)
+static void
+rhythmdb_query_model_cancel (RhythmDBModel *rmodel)
 {
-	g_signal_emit (G_OBJECT (model), rhythmdb_query_model_signals[COMPLETE], 0);
+	RhythmDBQueryModel *model = RHYTHMDB_QUERY_MODEL (rmodel);
+	rb_debug ("cancelling query");
+	g_async_queue_push (model->priv->query_complete, GINT_TO_POINTER (1));
+}
+
+void
+rhythmdb_query_model_signal_complete (RhythmDBQueryModel *model)
+{
+	struct RhythmDBQueryModelUpdate *update;
+
+	update = g_new0 (struct RhythmDBQueryModelUpdate, 1);
+	update->type = RHYTHMDB_QUERY_MODEL_UPDATE_QUERY_COMPLETE;
+
+	g_async_queue_push (model->priv->pending_updates, update);
+}
+
+void
+rhythmdb_query_model_set_connected (RhythmDBQueryModel *model, gboolean connected)
+{
+	model->priv->connected = connected;
+}
+
+void
+rhythmdb_query_model_finish_complete (RhythmDBQueryModel *model)
+{
+	if (!rb_thread_helpers_in_main_thread ())
+		g_async_queue_pop (model->priv->query_complete);
 }
 
 gboolean
@@ -493,18 +557,6 @@ rhythmdb_query_model_has_pending_changes (RhythmDBModel *rmodel)
 	return g_async_queue_length (model->priv->pending_updates) > 0;
 }
 
-static void
-rhythmdb_query_model_cancel (RhythmDBModel *model)
-{
-	rb_debug ("cancelling query");
-}
-
-static gboolean
-rhythmdb_query_model_sortable (RhythmDBModel *model)
-{
-	return TRUE;
-}
-
 static inline GSequencePtr
 choose_sequence_element (GSequence *seq)
 {
@@ -518,9 +570,10 @@ rhythmdb_query_model_entry_added_cb (RhythmDB *db, RhythmDBEntry *entry,
 				     RhythmDBQueryModel *model)
 {
 	if (G_LIKELY (model->priv->query)) {
-		if (model->priv->max_size > 0
-		    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_size)
+		if (model->priv->max_count > 0
+		    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_count)
 			return;
+		/* Check size later */
 		
 		if (rhythmdb_evaluate_query (db, model->priv->query, entry)) {
 			rhythmdb_query_model_add_entry (model, entry);
@@ -535,18 +588,18 @@ rhythmdb_query_model_entry_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
 	if (g_hash_table_lookup (model->priv->reverse_map, entry) != NULL) {
 		struct RhythmDBQueryModelUpdate *update;
 
-		rb_debug ("queueing entry deletion");
+		rb_debug ("queueing entry change");
 
 		update = g_new (struct RhythmDBQueryModelUpdate, 1);
 		update->type = RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED;
 		update->entry = entry;
-
+	
 		/* Called with a locked database */
 		rhythmdb_entry_ref_unlocked (model->priv->db, entry);
 
 		g_async_queue_push (model->priv->pending_updates, update);
 	} else {
-		/* the changed entrie may now satisfy the query so we test it */
+		/* the changed entry may now satisfy the query so we test it */
 		rhythmdb_query_model_entry_added_cb (db, entry, model);
 	}
 }
@@ -565,10 +618,16 @@ rhythmdb_query_model_add_entry (RhythmDBQueryModel *model, RhythmDBEntry *entry)
 {
 	struct RhythmDBQueryModelUpdate *update;
 
-	if (model->priv->max_size > 0
-	    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_size)
+	if (model->priv->max_count > 0
+	    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_count)
 		return;
-	
+	/* Check size later */
+
+	if (!model->priv->connected) {
+		rhythmdb_query_model_do_insert (model, entry, FALSE);
+		return;
+	}
+
 	update = g_new (struct RhythmDBQueryModelUpdate, 1);
 	update->type = RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED;
 	update->entry = entry;
@@ -588,6 +647,11 @@ rhythmdb_query_model_remove_entry (RhythmDBQueryModel *model, RhythmDBEntry *ent
 	if (g_hash_table_lookup (model->priv->reverse_map, entry) != NULL) {
 		struct RhythmDBQueryModelUpdate *update;
 
+		if (!model->priv->connected) {
+			rhythmdb_query_model_do_delete (model, entry, FALSE);
+			return;
+		}
+
 		update = g_new (struct RhythmDBQueryModelUpdate, 1);
 		update->type = RHYTHMDB_QUERY_MODEL_UPDATE_ROW_DELETED;
 		update->entry = entry;
@@ -599,7 +663,17 @@ rhythmdb_query_model_remove_entry (RhythmDBQueryModel *model, RhythmDBEntry *ent
 	}
 }
 
+GnomeVFSFileSize
+rhythmdb_query_model_get_size (RhythmDBQueryModel *model)
+{
+	return model->priv->total_size;
+}
 
+long
+rhythmdb_query_model_get_duration (RhythmDBQueryModel *model)
+{
+	return model->priv->total_duration;
+}
 
 static int
 compare_times (GTimeVal *a, GTimeVal *b)
@@ -616,6 +690,83 @@ compare_times (GTimeVal *a, GTimeVal *b)
 		return -1;
 }
 
+static GSequencePtr
+rhythmdb_query_model_do_insert (RhythmDBQueryModel *model,
+				RhythmDBEntry *entry,
+				gboolean lock)
+{
+	GSequencePtr ptr;
+	long size;
+	long duration;
+
+	/* we check again if the entry already exists in the hash table */
+	if (g_hash_table_lookup (model->priv->reverse_map, entry) != NULL)
+		return NULL;
+
+	if (model->priv->max_count > 0
+	    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_count)
+		return NULL;
+
+	if (lock)
+		rhythmdb_read_lock (model->priv->db);
+	size = rhythmdb_entry_get_long (model->priv->db, entry,
+					RHYTHMDB_PROP_FILE_SIZE);
+	duration = rhythmdb_entry_get_long (model->priv->db,
+					    entry, RHYTHMDB_PROP_DURATION);
+	if (lock)
+		rhythmdb_read_unlock (model->priv->db);
+
+	if (model->priv->max_size_mb > 0
+	    && ((model->priv->total_size + size) / (1024*1024)) >= model->priv->max_size_mb)
+		return NULL;
+
+	if (model->priv->sort_func) {
+		if (lock)
+			rhythmdb_read_lock (model->priv->db);
+		ptr = g_sequence_insert_sorted (model->priv->entries, entry,
+						model->priv->sort_func,
+						model->priv->sort_user_data);
+		if (lock)
+			rhythmdb_read_unlock (model->priv->db);
+	} else {
+		ptr = g_sequence_get_end_ptr (model->priv->entries);
+		g_sequence_insert (ptr, entry);
+		ptr = g_sequence_ptr_prev (ptr);
+	}
+
+	g_hash_table_insert (model->priv->reverse_map, entry, ptr);
+
+	model->priv->total_duration += duration;
+	model->priv->total_size += size;
+
+	return ptr;
+}
+
+static void
+rhythmdb_query_model_do_delete (RhythmDBQueryModel *model,
+				RhythmDBEntry *entry,
+				gboolean lock)
+{
+	GSequencePtr ptr;
+	
+	ptr = g_hash_table_lookup (model->priv->reverse_map, entry);
+
+	if (ptr == NULL)
+		return;
+
+	if (lock)
+		rhythmdb_read_lock (model->priv->db);
+	model->priv->total_duration -= rhythmdb_entry_get_long (model->priv->db, entry,
+								RHYTHMDB_PROP_DURATION);
+	model->priv->total_size -= rhythmdb_entry_get_long (model->priv->db, entry,
+							    RHYTHMDB_PROP_FILE_SIZE);
+	if (lock)
+		rhythmdb_read_unlock (model->priv->db);
+
+	g_sequence_remove (ptr);
+	g_hash_table_remove (model->priv->reverse_map, entry);
+}
+	
 /* Threading: main thread only, should hold GDK lock
  */
 static gboolean
@@ -627,6 +778,9 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 	struct RhythmDBQueryModelUpdate *update;
 	guint count = 0;
 
+	if (G_UNLIKELY (model->priv->cancelled))
+		return FALSE;
+
 	while ((update = g_async_queue_try_pop (model->priv->pending_updates)) != NULL) {
 		GtkTreePath *path;
 		GtkTreeIter iter;
@@ -641,36 +795,17 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 		switch (update->type) {
 		case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED:
 		{
-			/* we check again if the entry already exists in the hash table */
-			if (g_hash_table_lookup (model->priv->reverse_map, update->entry) != NULL)
-				break;
-			
-			if (model->priv->max_size > 0
-			    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_size)
-				break;
+			ptr = rhythmdb_query_model_do_insert (model, update->entry, TRUE);
 
-			if (model->priv->sort_func)
-				ptr = g_sequence_insert_sorted (model->priv->entries, update->entry,
-								model->priv->sort_func,
-								model->priv->sort_user_data);
-			else {
-				ptr = g_sequence_get_end_ptr (model->priv->entries);
-				g_sequence_insert (ptr, update->entry);
-				ptr = g_sequence_ptr_prev (ptr);
-			}
+			if (ptr == NULL)
+				break;
 
 			iter.user_data = ptr;
-			g_hash_table_insert (model->priv->reverse_map,
-					     update->entry,
-					     ptr);
-
 			path = rhythmdb_query_model_get_path (GTK_TREE_MODEL (model),
 							      &iter);
-
 			gtk_tree_model_row_inserted (GTK_TREE_MODEL (model),
 						     path, &iter);
 			gtk_tree_path_free (path);
-			break;
 		}
 		case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED:
 		{
@@ -681,10 +816,8 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 				break;
 
 			iter.user_data = ptr;
-
 			path = rhythmdb_query_model_get_path (GTK_TREE_MODEL (model),
 							      &iter);
-
 			rb_debug ("emitting row changed");
 			gtk_tree_model_row_changed (GTK_TREE_MODEL (model),
 						    path, &iter);
@@ -693,27 +826,26 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 		}
 		case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_DELETED:
 		{
-			GtkTreePath *path;
-
 			ptr = g_hash_table_lookup (model->priv->reverse_map,
 						   update->entry);
 
 			if (ptr == NULL)
 				break;
-
 			index = g_sequence_ptr_get_position (ptr);
 
 			path = gtk_tree_path_new ();
-
 			gtk_tree_path_append_index (path, index);
-				
 			rb_debug ("emitting row deleted");
 			gtk_tree_model_row_deleted (GTK_TREE_MODEL (model),
 						    path);
-
+			rhythmdb_query_model_do_delete (model, update->entry, TRUE);
+			
 			gtk_tree_path_free (path);
-			g_sequence_remove (ptr);
-			g_hash_table_remove (model->priv->reverse_map, update->entry);
+		}
+		case RHYTHMDB_QUERY_MODEL_UPDATE_QUERY_COMPLETE:
+		{
+			g_signal_emit (G_OBJECT (model), rhythmdb_query_model_signals[COMPLETE], 0);
+			g_async_queue_push (model->priv->query_complete, GINT_TO_POINTER (1));
 			break;
 		}
 		}
@@ -721,7 +853,7 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 		processed = g_list_prepend (processed, update);
 
 		count++;
-		if (timeout && count / 4 > 0) {
+		if (timeout && count / 8 > 0) {
 			/* Do this here at the bottom, so we do at least one update. */
 			g_get_current_time (&now);
 			if (compare_times (timeout,&now) < 0)
@@ -746,6 +878,8 @@ rhythmdb_query_model_poll (RhythmDBModel *rmodel, GTimeVal *timeout)
 			rhythmdb_entry_unref (model->priv->db, update->entry);
 			rhythmdb_entry_unref (model->priv->db, update->entry);
 			break;
+		case RHYTHMDB_QUERY_MODEL_UPDATE_QUERY_COMPLETE:
+			break;
 		}
 
 		g_free (update);
@@ -764,8 +898,11 @@ rhythmdb_query_model_entry_to_iter (RhythmDBModel *rmodel, RhythmDBEntry *entry,
 
 	ptr = g_hash_table_lookup (model->priv->reverse_map, entry);
 
-	if (ptr == NULL)
+	if (G_UNLIKELY (ptr == NULL)) {
+		// Invalidate iterator so future uses break quickly.
+		iter->stamp = !(model->priv->stamp);
 		return FALSE;
+	}
 
 	iter->stamp = model->priv->stamp;
 	iter->user_data = ptr;
@@ -773,22 +910,48 @@ rhythmdb_query_model_entry_to_iter (RhythmDBModel *rmodel, RhythmDBEntry *entry,
 }
 
 static gboolean
-rhythmdb_query_model_multi_row_draggable (EggTreeMultiDragSource *dragsource,
+rhythmdb_query_model_row_draggable (RbTreeDragSource *dragsource,
 					  GList *paths)
 {
 	return TRUE;
 }
 
 static gboolean
-rhythmdb_query_model_multi_drag_data_delete (EggTreeMultiDragSource *dragsource, GList *paths)
+rhythmdb_query_model_drag_data_delete (RbTreeDragSource *dragsource, GList *paths)
 {
 	RhythmDBQueryModel *model = RHYTHMDB_QUERY_MODEL (dragsource);
-	return model->priv->sort_func == NULL;
+	GtkTreeModel *treemodel = GTK_TREE_MODEL (model);
+
+	/* we don't delete if it is a reorder drag and drop because the deletion already
+	   occured in rhythmdb_query_model_drag_data_received */
+	if (model->priv->sort_func == NULL && !model->priv->reorder_drag_and_drop) {
+
+		RhythmDBEntry *entry;
+		GtkTreeIter iter;
+		GtkTreePath *path;
+
+		for (; paths; paths = paths->next) {
+
+			path = gtk_tree_row_reference_get_path (paths->data);
+
+			if (path) {
+				if (rhythmdb_query_model_get_iter (treemodel, &iter, path)) {
+					entry = g_sequence_ptr_get_data (iter.user_data);
+					rhythmdb_query_model_remove_entry (model, entry);
+				}
+				gtk_tree_path_free (path);
+			}
+		}
+	}
+
+	model->priv->reorder_drag_and_drop = FALSE;
+	return TRUE;
+
 }
 
 static gboolean
-rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource, 
-					  GList *paths, 
+rhythmdb_query_model_drag_data_get (RbTreeDragSource *dragsource,
+					  GList *paths,
 					  GtkSelectionData *selection_data)
 {
 	RhythmDBQueryModel *model = RHYTHMDB_QUERY_MODEL (dragsource);
@@ -807,11 +970,10 @@ rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource,
 		for (tem = paths; tem; tem = tem->next) {
 			GtkTreeIter iter;
 			GtkTreePath *path;
-			GSequencePtr ptr;
 			const char *location;
 
 			path = gtk_tree_row_reference_get_path (tem->data);
-	  
+
 			gtk_tree_model_get_iter (GTK_TREE_MODEL (model), &iter, path);
 
 			entry = g_sequence_ptr_get_data (iter.user_data);
@@ -823,19 +985,6 @@ rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource,
 
 			rhythmdb_read_unlock (model->priv->db);
 
-			if (!model->priv->sort_func) {
-				ptr = g_hash_table_lookup (model->priv->reverse_map,
-							   entry);
-
-				rb_debug ("emitting row deleted");
-				gtk_tree_model_row_deleted (GTK_TREE_MODEL (model),
-							    path);
-
-				gtk_tree_path_free (path);
-				g_sequence_remove (ptr);
-				g_hash_table_remove (model->priv->reverse_map, entry);
-			}
-
 			if (tem->next)
 				g_string_append (data, "\r\n");
 		}
@@ -846,7 +995,7 @@ rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource,
 					strlen (data->str));
 
 		g_string_free (data, TRUE);
-      
+
 		return TRUE;
 	}
 
@@ -854,11 +1003,12 @@ rhythmdb_query_model_multi_drag_data_get (EggTreeMultiDragSource *dragsource,
 }
 
 static gboolean
-rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
+rhythmdb_query_model_drag_data_received (RbTreeDragDest *drag_dest,
 					 GtkTreePath *dest,
 					 GtkSelectionData  *selection_data)
 {
 	RhythmDBQueryModel *model = RHYTHMDB_QUERY_MODEL (drag_dest);
+	GtkTreePath *path;
 
 	rb_debug ("drag received");
 
@@ -874,10 +1024,13 @@ rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
 
 		strv = g_strsplit (selection_data->data, "\r\n", -1);
 
-		g_assert (rhythmdb_query_model_get_iter (GTK_TREE_MODEL (model), &iter, dest));
-		ptr = iter.user_data;
+		if (rhythmdb_query_model_get_iter (GTK_TREE_MODEL (model), &iter, dest))
+			ptr = iter.user_data;
+		else
+			ptr = g_sequence_get_end_ptr (model->priv->entries);
 
-		while (strv[i]) {
+
+		for (; strv[i]; i++) {
 			GSequencePtr tem_ptr;
 			GtkTreeIter tem_iter;
 			GtkTreePath *tem_path;
@@ -891,6 +1044,36 @@ rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
 			if (entry == NULL)
 				rhythmdb_add_uri_async (model->priv->db, strv[i]);
 			else {
+				GSequencePtr old_ptr = g_hash_table_lookup (model->priv->reverse_map,
+									    entry);
+
+				/* the entry already exists it is either a reorder drag and drop
+				   (or a drag and drop from another application), so we delete
+				   the existing one before adding it again. */
+				if (old_ptr) {
+
+					model->priv->reorder_drag_and_drop = TRUE;
+
+					/* trying to drag drop an entry on itself ! */
+					if (old_ptr == ptr)
+						continue;
+
+					path = gtk_tree_path_new ();
+					gtk_tree_path_append_index (path,
+								    g_sequence_ptr_get_position (old_ptr));
+
+					rb_debug ("emitting row deleted");
+					gtk_tree_model_row_deleted (GTK_TREE_MODEL (model),
+								    path);
+					gtk_tree_path_free (path);
+					g_sequence_remove (old_ptr);
+					g_hash_table_remove (model->priv->reverse_map, entry);
+
+				} else {
+
+					model->priv->reorder_drag_and_drop = FALSE;
+				}
+
 				g_sequence_insert (ptr, entry);
 
 				tem_ptr = g_sequence_ptr_prev (ptr);
@@ -902,13 +1085,12 @@ rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
 
 				tem_path = rhythmdb_query_model_get_path (GTK_TREE_MODEL (model),
 									  &tem_iter);
-				
+
 				rb_debug ("emitting row inserted from dnd");
 				gtk_tree_model_row_inserted (GTK_TREE_MODEL (model),
 							     tem_path, &tem_iter);
 				gtk_tree_path_free (tem_path);
 			}
-			i++;
 		}
 
 		g_strfreev (strv);
@@ -918,7 +1100,7 @@ rhythmdb_query_model_drag_data_received (GtkTreeDragDest *drag_dest,
 }
 
 static gboolean
-rhythmdb_query_model_row_drop_possible (GtkTreeDragDest *drag_dest,
+rhythmdb_query_model_row_drop_possible (RbTreeDragDest *drag_dest,
 					GtkTreePath *dest,
 					GtkSelectionData  *selection_data)
 {
