@@ -1,7 +1,7 @@
 /*
  *  arch-tag: Implementation of RhythmDB - Rhythmbox backend queryable database
  *
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,6 +19,8 @@
  *
  */
 
+#include <config.h>
+
 #include "rhythmdb.h"
 #include "rhythmdb-model.h"
 #include "rhythmdb-legacy.h"
@@ -53,13 +55,18 @@ struct RhythmDBPrivate
 
 	GHashTable *legacy_id_map;
 
+	gboolean query_in_progress;
+
 	GAsyncQueue *main_thread_cancel;
 	GAsyncQueue *update_queue;
 	GAsyncQueue *add_queue;
+	GAsyncQueue *action_queue;
 
 	GHashTable *added_entries;
 	GHashTable *changed_entries;
 
+	GHashTable *propname_map;
+
 	guint thread_reaper_id;
 
 	GAsyncQueue *status_queue;
@@ -69,6 +76,16 @@ struct RhythmDBPrivate
 	gboolean exiting;
 };
 
+struct RhythmDBAction
+{
+	enum {
+		RHYTHMDB_ACTION_SET,
+	} type;
+	RhythmDBEntry *entry;
+	guint propid;
+	GValue value;
+};
+
 struct RhythmDBQueryThreadData
 {
 	GPtrArray *query;
@@ -91,6 +108,7 @@ static void rhythmdb_get_property (GObject *object,
 					GParamSpec *pspec);
 static gpointer add_thread_main (RhythmDB *db);
 static gpointer update_thread_main (RhythmDB *db);
+static gpointer action_thread_main (RhythmDB *db);
 static void update_song (RhythmDB *db, RhythmDBEntry *entry, GError **error);
 gboolean reap_dead_threads (RhythmDB *db);
 gpointer query_thread_main (struct RhythmDBQueryThreadData *data, RhythmDB *db);
@@ -270,6 +288,7 @@ rhythmdb_init (RhythmDB *db)
 	db->priv->main_thread_cancel = g_async_queue_new ();
 	db->priv->update_queue = g_async_queue_new ();
 	db->priv->add_queue = g_async_queue_new ();
+	db->priv->action_queue = g_async_queue_new ();
 
 	db->priv->status_queue = g_async_queue_new ();
 	
@@ -296,10 +315,17 @@ rhythmdb_init (RhythmDB *db)
 	g_type_class_unref (prop_class);
 	g_type_class_unref (unsaved_prop_class);
 
+	db->priv->propname_map = g_hash_table_new (g_str_hash, g_str_equal);
+
+	for (i = 0; i < RHYTHMDB_NUM_PROPERTIES; i++) {
+		const char *name = rhythmdb_nice_elt_name_from_propid (db, i);
+		g_hash_table_insert (db->priv->propname_map, (gpointer) name, GINT_TO_POINTER (i));
+	}
+
 	db->priv->legacy_id_map = g_hash_table_new (NULL, NULL);
 
 	db->priv->query_thread_pool = g_thread_pool_new ((GFunc) query_thread_main,
-							 db, 3, TRUE, NULL);
+							 db, 1, TRUE, NULL);
 	db->priv->thread_reaper_id
 		= g_idle_add ((GSourceFunc) reap_dead_threads, db);
 	g_async_queue_ref (db->priv->main_thread_cancel);
@@ -308,6 +334,9 @@ rhythmdb_init (RhythmDB *db)
 	g_async_queue_ref (db->priv->main_thread_cancel);
 	g_async_queue_ref (db->priv->update_queue);
 	g_thread_create ((GThreadFunc) update_thread_main, db, FALSE, NULL);
+	g_async_queue_ref (db->priv->main_thread_cancel);
+	g_async_queue_ref (db->priv->action_queue);
+	g_thread_create ((GThreadFunc) action_thread_main, db, FALSE, NULL);
 }
 
 gboolean
@@ -344,6 +373,8 @@ rhythmdb_shutdown (RhythmDB *db)
 		db->priv->outstanding_threads--;
 	}
 
+	/* Once for each thread */
+	g_async_queue_pop (db->priv->main_thread_cancel);
 	g_async_queue_pop (db->priv->main_thread_cancel);
 	g_async_queue_pop (db->priv->main_thread_cancel);
 }
@@ -368,6 +399,8 @@ rhythmdb_finalize (GObject *object)
 
 	g_free (db->priv->column_types);
 
+	g_hash_table_destroy (db->priv->propname_map);
+
 	g_free (db->priv->name);
 
 	g_free (db->priv);
@@ -495,7 +528,8 @@ rhythmdb_entry_new (RhythmDB *db, RhythmDBEntryType type, const char *uri)
 	if (!db->priv->changed_entries)
 		db->priv->added_entries = g_hash_table_new (NULL, NULL);
 
-	g_hash_table_insert (db->priv->added_entries, ret, NULL);
+	if (ret != NULL)
+		g_hash_table_insert (db->priv->added_entries, ret, NULL);
 	return ret;
 }
 
@@ -696,7 +730,6 @@ push_err (RhythmDB *db, const char *uri, GError *error)
 	g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) signal_err_idle, loaderr, NULL);
 }
 
-
 static gpointer
 read_queue (GAsyncQueue *queue, gboolean *cancel)
 {
@@ -735,6 +768,7 @@ add_thread_main (RhythmDB *db)
 		uri = read_queue (db->priv->add_queue, &db->priv->exiting);
 		if (!uri)
 			break;
+
 		realuri = rb_uri_resolve_symlink (uri);
 
 		if (rb_uri_is_directory (uri) == FALSE)
@@ -748,8 +782,6 @@ add_thread_main (RhythmDB *db)
 
 		g_free (realuri);
 		g_free (uri);
-			
-		g_usleep (10);
 	}
 
 	rb_debug ("exiting");
@@ -826,8 +858,6 @@ update_thread_main (RhythmDB *db)
 
 		if (error != NULL)
 			push_err (db, NULL, error);
-
-		g_usleep (10);
 	}
 
 	rb_debug ("exiting");
@@ -838,6 +868,38 @@ update_thread_main (RhythmDB *db)
 	return NULL;
 }
 
+static gpointer
+action_thread_main (RhythmDB *db)
+{
+	while (TRUE) {
+		struct RhythmDBAction *action;
+
+		action = read_queue (db->priv->action_queue, &db->priv->exiting);
+
+		if (action == NULL)
+			break;
+		
+		switch (action->type)
+		{
+		case RHYTHMDB_ACTION_SET:
+			rhythmdb_write_lock (db);
+			rhythmdb_entry_set (db, action->entry, action->propid,
+					    &action->value);
+			rhythmdb_write_unlock (db);
+			break;
+		}
+		g_value_unset (&action->value);
+		g_free (action);
+	}
+
+	rb_debug ("exiting");
+	g_async_queue_unref (db->priv->action_queue);
+	g_async_queue_push (db->priv->main_thread_cancel, GINT_TO_POINTER (1));
+	g_async_queue_unref (db->priv->main_thread_cancel);
+	g_thread_exit (NULL);
+	return NULL;
+}
+
 void
 rhythmdb_add_uri_async (RhythmDB *db, const char *uri)
 {
@@ -938,6 +1000,18 @@ rhythmdb_entry_set (RhythmDB *db, RhythmDBEntry *entry,
 	rhythmdb_entry_sync_mirrored (db, entry, propid, value);
 }
 
+void
+rhythmdb_entry_queue_set (RhythmDB *db, RhythmDBEntry *entry,
+			  guint propid, GValue *value)
+{
+	struct RhythmDBAction *action = g_new0 (struct RhythmDBAction, 1);
+	action->entry = entry;
+	action->propid = propid;
+	g_value_init (&action->value, G_VALUE_TYPE (value));
+	g_value_copy (value, &action->value);
+	g_async_queue_push (db->priv->action_queue, action);
+}
+
 void
 rhythmdb_entry_sync_mirrored (RhythmDB *db, RhythmDBEntry *entry, guint propid,
 			      GValue *value)
@@ -1065,6 +1139,7 @@ rhythmdb_query_parse_valist (RhythmDB *db, va_list args)
 			break;
 		case RHYTHMDB_QUERY_PROP_EQUALS:
 		case RHYTHMDB_QUERY_PROP_LIKE:
+		case RHYTHMDB_QUERY_PROP_NOT_LIKE:
 		case RHYTHMDB_QUERY_PROP_GREATER:
 		case RHYTHMDB_QUERY_PROP_LESS:
 			data->propid = va_arg (args, guint);
@@ -1134,6 +1209,7 @@ rhythmdb_query_free (GPtrArray *query)
 			break;
 		case RHYTHMDB_QUERY_PROP_EQUALS:
 		case RHYTHMDB_QUERY_PROP_LIKE:
+		case RHYTHMDB_QUERY_PROP_NOT_LIKE:
 		case RHYTHMDB_QUERY_PROP_GREATER:
 		case RHYTHMDB_QUERY_PROP_LESS:
 			g_value_unset (data->val);
@@ -1149,6 +1225,251 @@ rhythmdb_query_free (GPtrArray *query)
 	g_ptr_array_free (query, TRUE);
 }
 
+inline const char *
+rhythmdb_nice_elt_name_from_propid (RhythmDB *db, gint propid)
+{
+	switch (propid) {
+	case RHYTHMDB_PROP_TYPE:
+		return "type";
+	case RHYTHMDB_PROP_TITLE:
+		return "title";
+	case RHYTHMDB_PROP_TITLE_FOLDED:
+		return "title-folded";
+	case RHYTHMDB_PROP_TITLE_SORT_KEY:
+		return "title-sort-key";
+	case RHYTHMDB_PROP_GENRE:
+		return "genre";
+	case RHYTHMDB_PROP_GENRE_FOLDED:
+		return "genre-folded";
+	case RHYTHMDB_PROP_GENRE_SORT_KEY:
+		return "genre-sort-key";
+	case RHYTHMDB_PROP_ARTIST:
+		return "artist";
+	case RHYTHMDB_PROP_ARTIST_FOLDED:
+		return "artist-folded";
+	case RHYTHMDB_PROP_ARTIST_SORT_KEY:
+		return "artist-sort-key";
+	case RHYTHMDB_PROP_ALBUM:
+		return "album";
+	case RHYTHMDB_PROP_ALBUM_FOLDED:
+		return "album-folded";
+	case RHYTHMDB_PROP_ALBUM_SORT_KEY:
+		return "album-sort-key";
+	case RHYTHMDB_PROP_TRACK_NUMBER:
+		return "track-number";
+	case RHYTHMDB_PROP_DURATION:
+		return "duration";
+	case RHYTHMDB_PROP_FILE_SIZE:
+		return "file-size";
+	case RHYTHMDB_PROP_LOCATION:
+		return "location";
+	case RHYTHMDB_PROP_MTIME:
+		return "mtime";
+	case RHYTHMDB_PROP_RATING:
+		return "rating";
+	case RHYTHMDB_PROP_PLAY_COUNT:
+		return "play-count";
+	case RHYTHMDB_PROP_LAST_PLAYED:
+		return "last-played";
+	case RHYTHMDB_PROP_LAST_PLAYED_STR:
+		return "last-played-str";
+	case RHYTHMDB_PROP_QUALITY:
+		return "quality";
+	default:
+		g_assert_not_reached ();
+	}
+	return NULL;
+}
+
+inline int
+rhythmdb_propid_from_nice_elt_name (RhythmDB *db, const char *name)
+{
+	gpointer ret, orig;	
+	if (g_hash_table_lookup_extended (db->priv->propname_map, name,
+					  &orig, &ret)) {
+		return GPOINTER_TO_INT (ret);
+	}
+	return -1;
+}
+
+static void
+write_encoded_gvalue (xmlNodePtr node,
+		      GValue *val)
+{
+	char *strval;
+	char *quoted;
+
+	switch (G_VALUE_TYPE (val))
+	{
+	case G_TYPE_STRING:
+		strval = g_value_dup_string (val);
+		break;
+	case G_TYPE_BOOLEAN:
+		strval = g_strdup_printf ("%d", g_value_get_boolean (val));
+		break;
+	case G_TYPE_INT:
+		strval = g_strdup_printf ("%d", g_value_get_int (val));
+		break;
+	case G_TYPE_LONG:
+		strval = g_strdup_printf ("%ld", g_value_get_long (val));
+		break;
+	case G_TYPE_FLOAT:
+		strval = g_strdup_printf ("%f", g_value_get_float (val));
+		break;
+	case G_TYPE_DOUBLE:
+		strval = g_strdup_printf ("%f", g_value_get_double (val));
+		break;
+	default:
+		g_assert_not_reached ();
+		break;
+	}
+
+	quoted = xmlEncodeEntitiesReentrant (NULL, strval);
+	g_free (strval);
+	xmlNodeSetContent (node, quoted);
+	g_free (quoted);
+}
+
+static void
+read_encoded_property (RhythmDB *db,
+		       xmlNodePtr node,
+		       guint propid,
+		       GValue *val)
+{
+	char *content;
+	
+	g_value_init (val, db->priv->column_types[propid]);
+
+	content = xmlNodeGetContent (node);
+	
+	switch (G_VALUE_TYPE (val))
+	{
+	case G_TYPE_STRING:
+		g_value_set_string (val, content);
+		break;
+	case G_TYPE_BOOLEAN:
+		g_value_set_boolean (val, g_ascii_strtoull (content, NULL, 10));
+		break;
+	case G_TYPE_INT:
+		g_value_set_int (val, g_ascii_strtoull (content, NULL, 10));
+		break;
+	case G_TYPE_LONG:
+		g_value_set_long (val, g_ascii_strtoull (content, NULL, 10));
+		break;
+	case G_TYPE_FLOAT:
+		g_value_set_float (val, g_ascii_strtod (content, NULL));
+		break;
+	case G_TYPE_DOUBLE:
+		g_value_set_float (val, g_ascii_strtod (content, NULL));
+		break;
+	default:
+		g_assert_not_reached ();
+		break;
+	}
+
+	g_free (content);
+}
+
+void
+rhythmdb_query_serialize (RhythmDB *db, GPtrArray *query,
+			  xmlNodePtr parent)
+{
+	guint i;
+	xmlNodePtr node = xmlNewChild (parent, NULL, "conjunction", NULL);
+	xmlNodePtr subnode;
+
+	for (i = 0; i < query->len; i++) {
+		RhythmDBQueryData *data = g_ptr_array_index (query, i);
+		
+		switch (data->type) {
+		case RHYTHMDB_QUERY_SUBQUERY:
+			subnode = xmlNewChild (node, NULL, "subquery", NULL);
+			rhythmdb_query_serialize (db, data->subquery, subnode);
+			break;
+		case RHYTHMDB_QUERY_PROP_LIKE:
+			subnode = xmlNewChild (node, NULL, "like", NULL);
+			xmlSetProp (subnode, "prop", rhythmdb_nice_elt_name_from_propid (db, data->propid));
+			write_encoded_gvalue (subnode, data->val);
+			break;
+		case RHYTHMDB_QUERY_PROP_NOT_LIKE:
+			subnode = xmlNewChild (node, NULL, "not-like", NULL);
+			xmlSetProp (subnode, "prop", rhythmdb_nice_elt_name_from_propid (db, data->propid));
+			write_encoded_gvalue (subnode, data->val);
+			break;
+		case RHYTHMDB_QUERY_PROP_EQUALS:
+			subnode = xmlNewChild (node, NULL, "equals", NULL);
+			xmlSetProp (subnode, "prop", rhythmdb_nice_elt_name_from_propid (db, data->propid));
+			write_encoded_gvalue (subnode, data->val);
+			break;
+		case RHYTHMDB_QUERY_DISJUNCTION:
+			subnode = xmlNewChild (node, NULL, "disjunction", NULL);
+			break;
+		case RHYTHMDB_QUERY_END:
+			break;
+		case RHYTHMDB_QUERY_PROP_GREATER:
+		case RHYTHMDB_QUERY_PROP_LESS:
+			g_assert_not_reached ();
+			break;
+		}		
+	}
+}
+
+GPtrArray *
+rhythmdb_query_deserialize (RhythmDB *db, xmlNodePtr parent)
+{
+	GPtrArray *query = g_ptr_array_new ();
+	xmlNodePtr child;
+
+	g_assert (!strcmp (parent->name, "conjunction"));
+	
+	for (child = parent->children; child; child = child->next) {
+		RhythmDBQueryData *data;
+
+		if (xmlNodeIsText (child))
+			continue;
+
+		data = g_new0 (RhythmDBQueryData, 1);
+
+		if (!strcmp (child->name, "subquery")) {
+			xmlNodePtr subquery;
+			data->type = RHYTHMDB_QUERY_SUBQUERY;
+			subquery = child->children;
+			while (xmlNodeIsText (subquery))
+				subquery = subquery->next;
+			
+			data->subquery = rhythmdb_query_deserialize (db, subquery);
+		} else if (!strcmp (child->name, "disjunction")) {
+			data->type = RHYTHMDB_QUERY_DISJUNCTION;
+		} else if (!strcmp (child->name, "like")) {
+			data->type = RHYTHMDB_QUERY_PROP_LIKE;
+		} else if (!strcmp (child->name, "not-like")) {
+			data->type = RHYTHMDB_QUERY_PROP_NOT_LIKE;
+		} else if (!strcmp (child->name, "equals")) {
+			data->type = RHYTHMDB_QUERY_PROP_EQUALS;
+		} else
+ 			g_assert_not_reached ();
+
+		if (!strcmp (child->name, "like")
+		    || !strcmp (child->name, "not-like")
+		    || !strcmp (child->name, "equals")) {
+			char *propstr = xmlGetProp (child, "prop");
+			gint propid = rhythmdb_propid_from_nice_elt_name (db, propstr);
+			g_free (propstr);
+
+			g_assert (propid >= 0 && propid < RHYTHMDB_NUM_PROPERTIES);
+
+			data->propid = propid;
+			data->val = g_new0 (GValue, 1);
+
+			read_encoded_property (db, child, data->propid, data->val);
+		} 
+
+		g_ptr_array_add (query, data);
+	}
+
+	return query;
+}
+
 RhythmDBEntry *
 rhythmdb_entry_lookup_by_location (RhythmDB *db, const char *uri)
 {
@@ -1184,11 +1505,12 @@ query_thread_main (struct RhythmDBQueryThreadData *data,
 				   data->main_model,
 				   &data->cancel);
 
-	if (data->lock)
-		rhythmdb_read_unlock (db);
 
 	rb_debug ("completed");
-	rhythmdb_query_model_complete (RHYTHMDB_QUERY_MODEL (data->main_model));
+	rhythmdb_query_model_signal_complete (RHYTHMDB_QUERY_MODEL (data->main_model));
+	if (data->lock)
+		rhythmdb_read_unlock (db);
+	rhythmdb_query_model_finish_complete (RHYTHMDB_QUERY_MODEL (data->main_model));
 
 	rhythmdb_query_free (data->query);
 	g_async_queue_push (db->priv->status_queue, data->main_model);
@@ -1309,6 +1631,7 @@ rhythmdb_query_get_type (void)
 			ENUM_ENTRY (RHYTHMDB_QUERY_SUBQUERY, "Subquery"),
 			ENUM_ENTRY (RHYTHMDB_QUERY_PROP_EQUALS, "Property equivalence"),
 			ENUM_ENTRY (RHYTHMDB_QUERY_PROP_LIKE, "Fuzzy property matching"),
+			ENUM_ENTRY (RHYTHMDB_QUERY_PROP_NOT_LIKE, "Inverted fuzzy property matching"),
 			ENUM_ENTRY (RHYTHMDB_QUERY_PROP_GREATER, "True iff property1 > property2"),
 			ENUM_ENTRY (RHYTHMDB_QUERY_PROP_LESS, "True iff property1 < property2"),
 			{ 0, 0, 0 }
