@@ -2,7 +2,7 @@
  *  arch-tag: The Rhythmbox main entrypoint
  *
  *  Copyright (C) 2002 Jorn Baayen
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -60,6 +60,7 @@ static CORBA_Environment ev;
 static gboolean debug           = FALSE;
 static gboolean quit            = FALSE;
 static gboolean no_registration = FALSE;
+static gboolean dry_run		= FALSE;
 static gboolean print_playing = FALSE;
 static gboolean print_playing_path = FALSE;
 static gboolean playpause       = FALSE;
@@ -92,6 +93,7 @@ main (int argc, char **argv)
 		
 		{ "debug",           'd',  POPT_ARG_NONE,          &debug,                                        0, N_("Enable debugging code"),     NULL },
 		{ "no-registration", 'n',  POPT_ARG_NONE,          &no_registration,                              0, N_("Do not register the shell"), NULL },
+		{ "dry-run", 0,  POPT_ARG_NONE,          &dry_run,                             0, N_("Don't save any data permanently (implies --no-registration)"), NULL },
 		{ "quit",            'q',  POPT_ARG_NONE,          &quit,                                         0, N_("Quit Rhythmbox"),            NULL },
 		{ NULL,              '\0', POPT_ARG_INCLUDE_TABLE, (poptOption *) monkey_media_get_popt_table (), 0, N_("MonkeyMedia options:"),      NULL },
 		POPT_TABLEEND
@@ -133,7 +135,7 @@ main (int argc, char **argv)
 
 	rb_file_helpers_init ();
 
-	if (no_registration == FALSE) {
+	if (no_registration == FALSE && dry_run == FALSE) {
 		object = bonobo_activation_activate_from_id (RB_SHELL_OAFIID,
 			 				     Bonobo_ACTIVATION_FLAG_EXISTING_ONLY,
 							     NULL, NULL);
@@ -158,6 +160,8 @@ main (int argc, char **argv)
 
 			g_object_set_data (G_OBJECT (rb_shell), "rb-shell-no-registration",
 					   GINT_TO_POINTER (no_registration));
+			g_object_set_data (G_OBJECT (rb_shell), "rb-shell-dry-run",
+					   GINT_TO_POINTER (dry_run));
 			
 			g_idle_add ((GSourceFunc) rb_init, rb_shell);
 		}
@@ -211,7 +215,8 @@ rb_init (RBShell *shell)
 	argv = (char **) g_object_get_data (G_OBJECT (shell), "argv");
 	argc = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (shell), "argc"));
 
-	if (!g_object_get_data (G_OBJECT (shell), "rb-shell-no-registration"))
+	if (!g_object_get_data (G_OBJECT (shell), "rb-shell-no-registration")
+	    && (!g_object_get_data (G_OBJECT (shell), "rb-shell-dry-run")))
 		rb_handle_cmdline (argv, argc, FALSE);
 	
 	GDK_THREADS_LEAVE ();
