@@ -1,7 +1,7 @@
 /*
  *  arch-tag: Implementation of Rhythmbox playlist management object
  *
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -86,6 +86,8 @@ struct RBPlaylistManagerPrivate
 	GList *playlists;
 
 	RBPlaylistSource *loading_playlist;
+
+	char *firsturi;
 };
 
 enum
@@ -393,6 +395,34 @@ rb_playlist_manager_new (BonoboUIComponent *component, GtkWindow *window,
 	return mgr;
 }
 
+const char *
+rb_playlist_manager_parse_file (RBPlaylistManager *mgr, const char *uri)
+{
+	rb_debug ("loading playlist from %s", uri);
+
+	g_free (mgr->priv->firsturi);
+	mgr->priv->firsturi = NULL;
+
+	g_signal_emit (G_OBJECT (mgr), rb_playlist_manager_signals[PLAYLIST_LOAD_START], 0);
+
+	{
+		RBPlaylist *parser = rb_playlist_new ();
+
+		g_signal_connect (G_OBJECT (parser), "entry",
+				  G_CALLBACK (handle_playlist_entry_into_playlist_cb),
+				  mgr);
+
+		if (!rb_playlist_parse (parser, uri))
+			rb_error_dialog (_("Couldn't parse playlist"));
+		mgr->priv->loading_playlist = NULL;
+
+		g_object_unref (G_OBJECT (parser));
+	}
+
+	g_signal_emit (G_OBJECT (mgr), rb_playlist_manager_signals[PLAYLIST_LOAD_FINISH], 0);
+	return mgr->priv->firsturi;
+}
+
 static void
 rb_playlist_manager_source_deleted_cb (RBSource *source, RBPlaylistManager *mgr)
 {
@@ -606,6 +636,8 @@ rb_playlist_manager_cmd_new_automatic_playlist (BonoboUIComponent *component,
 {
 	RBQueryCreator *creator = RB_QUERY_CREATOR (rb_query_creator_new (mgr->priv->db));
 	RBSource *playlist;
+	RBQueryCreatorLimitType type;
+	guint limit, limit_count = 0, limit_size = 0;
 	
 	switch (gtk_dialog_run (GTK_DIALOG (creator)))
 	{
@@ -615,10 +647,15 @@ rb_playlist_manager_cmd_new_automatic_playlist (BonoboUIComponent *component,
 		return;
 	}
 
+	rb_query_creator_get_limit (creator, &type, &limit);
+	if (type == RB_QUERY_CREATOR_LIMIT_COUNT)
+		limit_count = limit;
+	if (type == RB_QUERY_CREATOR_LIMIT_MB)
+		limit_size = limit;
 	playlist = rb_playlist_manager_new_playlist (mgr, TRUE);
 	rb_playlist_source_set_query (RB_PLAYLIST_SOURCE (playlist),
 				      rb_query_creator_get_query (creator),
-				      rb_query_creator_get_limit (creator));
+				      limit_count, limit_size);
 	gtk_widget_destroy (GTK_WIDGET (creator));	
 }
 
@@ -654,27 +691,8 @@ load_playlist_response_cb (GtkDialog *dialog,
 	escaped_file = gnome_vfs_get_uri_from_local_path (file);
 	g_free (file);
 
-	rb_debug ("loading playlist from %s", escaped_file);
-
-	g_signal_emit (G_OBJECT (mgr), rb_playlist_manager_signals[PLAYLIST_LOAD_START], 0);
-
-	{
-		RBPlaylist *parser = rb_playlist_new ();
-		RBSource *playlist = rb_playlist_manager_new_playlist (mgr, FALSE);
-
-		g_signal_connect (G_OBJECT (parser), "entry",
-				  G_CALLBACK (handle_playlist_entry_into_playlist_cb),
-				  mgr);
-
-		mgr->priv->loading_playlist = RB_PLAYLIST_SOURCE (playlist);
-		if (!rb_playlist_parse (parser, escaped_file))
-			rb_error_dialog (_("Couldn't parse playlist"));
-		mgr->priv->loading_playlist = NULL;
-
-		g_object_unref (G_OBJECT (parser));
-	}
-
-	g_signal_emit (G_OBJECT (mgr), rb_playlist_manager_signals[PLAYLIST_LOAD_FINISH], 0);
+	rb_playlist_manager_parse_file (mgr, escaped_file);
+	g_free (escaped_file);
 }
 
 static void
@@ -773,5 +791,17 @@ static void
 handle_playlist_entry_into_playlist_cb (RBPlaylist *playlist, const char *uri, const char *title,
 					const char *genre, RBPlaylistManager *mgr)
 {
+	if (!mgr->priv->firsturi)
+		mgr->priv->firsturi = g_strdup (uri);
+	if (rb_uri_is_iradio (uri)) {
+		rb_iradio_source_add_station (mgr->priv->iradio_source,
+					      uri, title, genre);
+		return;
+	}
+
+	if (!mgr->priv->loading_playlist) {
+		mgr->priv->loading_playlist =
+			RB_PLAYLIST_SOURCE (rb_playlist_manager_new_playlist (mgr, FALSE));
+	}
 	add_uri_to_playlist (mgr, mgr->priv->loading_playlist, uri, title);
 }
