@@ -2,7 +2,7 @@
  *  arch-tag: Implementation of main Rhythmbox shell
  *
  *  Copyright (C) 2002, 2003 Jorn Baayen
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -94,6 +94,12 @@ static void rb_shell_corba_grab_focus (PortableServer_Servant _servant,
 				       CORBA_Environment *ev);
 static void rb_shell_corba_playpause (PortableServer_Servant _servant,
 					       CORBA_Environment *ev);
+static void rb_shell_corba_select (PortableServer_Servant _servant,
+				   const CORBA_char *uri,
+				   CORBA_Environment *ev);
+static void rb_shell_corba_play (PortableServer_Servant _servant,
+				 const CORBA_char *uri,
+				 CORBA_Environment *ev);
 static void rb_shell_corba_next (PortableServer_Servant _servant,
 					       CORBA_Environment *ev);
 static void rb_shell_corba_previous (PortableServer_Servant _servant,
@@ -104,9 +110,9 @@ static void rb_shell_corba_set_playing_time (PortableServer_Servant _servant,
 						   CORBA_long time, CORBA_Environment *ev);
 
 static Bonobo_PropertyBag rb_shell_corba_get_player_properties (PortableServer_Servant _servant, CORBA_Environment *ev);
+static void rb_shell_db_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
+				    RBShell *shell);
 
-void rb_shell_handle_playlist_entry (RBShell *shell, GList *locations, const char *title,
-				     const char *genre);
 static gboolean rb_shell_window_state_cb (GtkWidget *widget,
 					  GdkEvent *event,
 					  RBShell *shell);
@@ -173,6 +179,10 @@ static void rb_shell_cmd_current_song (BonoboUIComponent *component,
 				       RBShell *shell,
 				       const char *verbname);
 static void rb_shell_jump_to_current (RBShell *shell);
+static void rb_shell_jump_to_entry (RBShell *shell, RhythmDBEntry *entry);
+static void rb_shell_jump_to_entry_with_source (RBShell *shell, RBSource *source,
+						RhythmDBEntry *entry);
+static void rb_shell_play_entry (RBShell *shell, RhythmDBEntry *entry);
 static void rb_shell_quit (RBShell *shell);
 static void rb_shell_view_sourcelist_changed_cb (BonoboUIComponent *component,
 						 const char *path,
@@ -239,9 +249,13 @@ struct RBShellPrivate
 	GtkWidget *paned;
 	GtkWidget *sourcelist;
 	GtkWidget *notebook;
+	GtkWidget *hsep;
 
 	GList *sources;
 
+	gboolean db_dirty;
+	guint async_state_save_id;
+
 	RhythmDB *db;
 
 	RBShellPlayer *player_shell;
@@ -344,6 +358,8 @@ rb_shell_class_init (RBShellClass *klass)
 	epv->addToLibrary = rb_shell_corba_add_to_library;
 	epv->grabFocus    = rb_shell_corba_grab_focus;
 	epv->playPause = rb_shell_corba_playpause;
+	epv->select = rb_shell_corba_select;
+	epv->play = rb_shell_corba_play;
 	epv->previous = rb_shell_corba_previous;
 	epv->next = rb_shell_corba_next;
 	epv->getPlayingTime = rb_shell_corba_get_playing_time;
@@ -370,6 +386,40 @@ rb_shell_init (RBShell *shell)
 
 }
 
+static gboolean
+rb_shell_sync_state (RBShell *shell)
+{
+	if (g_object_get_data (G_OBJECT (shell), "rb-shell-dry-run"))
+		return FALSE;
+	
+	rb_debug ("saving playlists");
+	rb_playlist_manager_save_playlists (shell->priv->playlist_manager);
+
+	rb_debug ("saving db");
+	rhythmdb_read_lock (shell->priv->db);
+	rhythmdb_save (shell->priv->db);
+	rhythmdb_read_unlock (shell->priv->db);
+	return FALSE;
+}
+
+static gboolean
+idle_save_state (RBShell *shell)
+{
+	GDK_THREADS_ENTER ();
+
+	rb_shell_sync_state (shell);
+
+	GDK_THREADS_LEAVE ();
+	return FALSE;
+}
+
+static void
+rb_shell_db_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
+			RBShell *shell)
+{
+	shell->priv->db_dirty = TRUE;
+}
+
 static void
 rb_shell_finalize (GObject *object)
 {
@@ -395,7 +445,6 @@ rb_shell_finalize (GObject *object)
 	
 	g_list_free (shell->priv->sources);
 
-	rb_playlist_manager_save_playlists (shell->priv->playlist_manager);
 	g_object_unref (G_OBJECT (shell->priv->playlist_manager));
 
 	g_object_unref (G_OBJECT (shell->priv->clipboard_shell));
@@ -405,11 +454,6 @@ rb_shell_finalize (GObject *object)
 	rb_debug ("shutting down DB");
 	rhythmdb_shutdown (shell->priv->db);
 
-	rb_debug ("saving db");
-	rhythmdb_read_lock (shell->priv->db);
-	rhythmdb_save (shell->priv->db);
-	rhythmdb_read_unlock (shell->priv->db);
-
 	rb_debug ("unreffing DB");
 	g_object_unref (G_OBJECT (shell->priv->db));
 
@@ -436,6 +480,73 @@ rb_shell_new (void)
 	return s;
 }
 
+struct RBShellAction
+{
+	RBShell *shell;
+	char *uri;
+	guint tries;
+	enum {
+		RB_SHELL_ACTION_JUMP,
+		RB_SHELL_ACTION_PLAY,
+	} type;
+};
+
+static gboolean
+idle_do_action (struct RBShellAction *data)
+{
+	RhythmDBEntry *entry;
+	char *unquoted = NULL;
+	
+	GDK_THREADS_ENTER ();
+
+	data->tries++;
+
+	rhythmdb_read_lock (data->shell->priv->db);
+	entry = rhythmdb_entry_lookup_by_location (data->shell->priv->db, data->uri);
+	rhythmdb_read_unlock (data->shell->priv->db);
+
+	if (entry) {
+		switch (data->type) {
+		case RB_SHELL_ACTION_JUMP:
+			rb_shell_jump_to_entry (data->shell, entry);
+			break;
+		case RB_SHELL_ACTION_PLAY:
+			rb_shell_play_entry (data->shell, entry);
+			break;
+		}
+	} else if (data->tries < 4)
+		g_timeout_add (500 + data->tries*200, (GSourceFunc) idle_do_action, data);
+	else
+		g_warning ("No entry %s in db", data->uri);
+
+	GDK_THREADS_LEAVE ();
+
+	g_free (unquoted);
+	g_free (data->uri);
+	g_free (data);
+	return FALSE;
+}
+
+static void
+rb_shell_queue_jump (RBShell *shell, const char *uri)
+{
+	struct RBShellAction *data = g_new0 (struct RBShellAction, 1);
+	data->shell = shell;
+	data->uri = g_strdup (uri);
+	data->type = RB_SHELL_ACTION_JUMP;
+	g_idle_add ((GSourceFunc) idle_do_action, data);
+}
+
+static void
+rb_shell_queue_play (RBShell *shell, const char *uri)
+{
+	struct RBShellAction *data = g_new0 (struct RBShellAction, 1);
+	data->shell = shell;
+	data->uri = g_strdup (uri);
+	data->type = RB_SHELL_ACTION_PLAY;
+	g_idle_add ((GSourceFunc) idle_do_action, data);
+}
+
 static void
 rb_shell_corba_quit (PortableServer_Servant _servant,
                      CORBA_Environment *ev)
@@ -449,20 +560,6 @@ rb_shell_corba_quit (PortableServer_Servant _servant,
 	GDK_THREADS_LEAVE ();
 }
 
-static void
-handle_playlist_entry_cb (RBPlaylist *playlist, const char *uri, const char *title,
-			  const char *genre, RBShell *shell)
-{
-	/* We assume all HTTP is iradio.  This is probably a broken assumption,
-	 * but it's very difficult to really fix...
-	 */
-	if (rb_uri_is_iradio (uri) != FALSE) {
-		rb_iradio_source_add_station (shell->priv->iradio_source, uri, title, genre);
-	} else {
-		rhythmdb_add_uri_async (shell->priv->db, (char *) uri);
-	}
-}
-
 static void
 rb_shell_corba_handle_file (PortableServer_Servant _servant,
 			    const CORBA_char *uri,
@@ -478,14 +575,12 @@ rb_shell_corba_handle_file (PortableServer_Servant _servant,
 	}
 
 	parser = rb_playlist_new ();	
-	g_signal_connect (G_OBJECT (parser), "entry",
-			  G_CALLBACK (handle_playlist_entry_cb), shell);
-
-	/* Try parsing it as a playlist; otherwise just try adding it to
-	 * the library.
-	 */
-	if (!rb_playlist_parse (parser, uri))
+	if (rb_playlist_can_handle (parser, uri)) {
+		uri = rb_playlist_manager_parse_file (shell->priv->playlist_manager, uri);
+	} else 
 		rhythmdb_add_uri_async (shell->priv->db, uri);
+
+	rb_shell_queue_play (shell, uri);
 	g_object_unref (G_OBJECT (parser));
 }
 
@@ -546,9 +641,11 @@ get_song_info_from_player (RBShell *shell)
 	song_info = GNOME_Rhythmbox_SongInfo__alloc ();
 	rhythmdb_read_lock (db);
 	song_info->title = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_TITLE));
-	song_info->artist = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_ARTIST));
+	if (rb_source_have_artist_album (shell->priv->selected_source)) {
+		song_info->artist = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_ARTIST));
+		song_info->album = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_ALBUM));
+	}
 	song_info->genre = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_GENRE));
-	song_info->album = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_ALBUM));
 	song_info->path = CORBA_string_dup (rhythmdb_entry_get_string (db, entry, RHYTHMDB_PROP_LOCATION));
 	song_info->track_number = rhythmdb_entry_get_int (db, entry, RHYTHMDB_PROP_TRACK_NUMBER);
 	song_info->duration = rhythmdb_entry_get_long (db, entry, RHYTHMDB_PROP_DURATION);
@@ -676,6 +773,24 @@ rb_shell_corba_playpause (PortableServer_Servant _servant,
 	rb_shell_player_playpause (shell->priv->player_shell);
 }
 
+static void
+rb_shell_corba_select (PortableServer_Servant _servant,
+		       const CORBA_char *uri,
+		       CORBA_Environment *ev)
+{
+	RBShell *shell = RB_SHELL (bonobo_object (_servant));
+	rb_shell_queue_jump (shell, uri);
+}
+
+static void
+rb_shell_corba_play (PortableServer_Servant _servant,
+		     const CORBA_char *uri,
+		     CORBA_Environment *ev)
+{
+	RBShell *shell = RB_SHELL (bonobo_object (_servant));
+	rb_shell_queue_play (shell, uri);
+}
+
 static void
 rb_shell_corba_next (PortableServer_Servant _servant,
 		     CORBA_Environment *ev)
@@ -807,6 +922,18 @@ rb_shell_construct (RBShell *shell)
 			rhythmdb_load (shell->priv->db);
 		}
 	}
+	g_signal_connect_object (G_OBJECT (shell->priv->db),
+				 "entry_added",
+				 G_CALLBACK (rb_shell_db_changed_cb),
+				 shell, 0);
+	g_signal_connect_object (G_OBJECT (shell->priv->db),
+				 "entry_changed",
+				 G_CALLBACK (rb_shell_db_changed_cb),
+				 shell, 0);
+	g_signal_connect_object (G_OBJECT (shell->priv->db),
+				 "entry_deleted",
+				 G_CALLBACK (rb_shell_db_changed_cb),
+				 shell, 0);
 
 	if (!rhythmdb_exists && eel_gconf_get_boolean (CONF_FIRST_TIME)) {
 		rb_debug ("loading legacy library db");
@@ -903,8 +1030,9 @@ rb_shell_construct (RBShell *shell)
 
 	vbox = gtk_vbox_new (FALSE, 5);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);
-	gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (shell->priv->player_shell), FALSE, TRUE, 0);
-	gtk_box_pack_start (GTK_BOX (vbox), gtk_hseparator_new (), FALSE, FALSE, 0);
+ 	gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (shell->priv->player_shell), FALSE, TRUE, 0);	
+ 	shell->priv->hsep = gtk_hseparator_new ();
+ 	gtk_box_pack_start (GTK_BOX (vbox), shell->priv->hsep, FALSE, FALSE, 0);
 	gtk_box_pack_start (GTK_BOX (vbox), shell->priv->paned, TRUE, TRUE, 0);
 	gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (shell->priv->statusbar), FALSE, TRUE, 0);
 
@@ -990,7 +1118,8 @@ rb_shell_construct (RBShell *shell)
 		rb_debug ("No AudioCD device is available!");
 #endif
 	
-	if (!g_object_get_data (G_OBJECT (shell), "rb-shell-no-registration")) {
+	if (!g_object_get_data (G_OBJECT (shell), "rb-shell-no-registration")
+	    && !g_object_get_data (G_OBJECT (shell), "rb-shell-dry-run")) {
 		/* register with CORBA */
 		CORBA_exception_init (&ev);
 		
@@ -1028,6 +1157,8 @@ rb_shell_construct (RBShell *shell)
 		gtk_widget_hide (GTK_WIDGET (shell->priv->window));
 		rb_druid_show (druid);
 		g_object_unref (G_OBJECT (druid));
+
+		g_timeout_add (5000, (GSourceFunc) idle_save_state, shell);
 	}
 	
 	rb_statusbar_sync_state (shell->priv->statusbar);
@@ -1337,10 +1468,10 @@ rb_shell_set_duration (RBShell *shell, const char *duration)
 	if (shell->priv->cached_title == NULL) 
 		tooltip = g_strdup (_("Not playing"));
 	else if (!playing) {
-		tooltip = g_strdup_printf (_("%s\nPaused, %s remaining"),
+		tooltip = g_strdup_printf (_("%s\nPaused, %s"),
 					 shell->priv->cached_title, duration);
 	} else {
-		tooltip = g_strdup_printf (_("%s\n%s remaining"),
+		tooltip = g_strdup_printf (_("%s\n%s"),
 					   shell->priv->cached_title, duration);
 	}
 	
@@ -1649,6 +1780,7 @@ rb_shell_quit (RBShell *shell)
 {
 	rb_debug ("Quitting");
 
+	rb_shell_sync_state (shell);
 	bonobo_object_unref (BONOBO_OBJECT (shell));
 }
 
@@ -1698,11 +1830,15 @@ rb_shell_sync_smalldisplay (RBShell *shell)
 					CMD_PATH_VIEW_SOURCELIST, FALSE);
   
 		gtk_widget_hide (GTK_WIDGET (shell->priv->paned));
+ 		gtk_widget_hide (GTK_WIDGET (shell->priv->statusbar));
+ 		gtk_widget_hide (GTK_WIDGET (shell->priv->hsep));		
 	} else {
 		rb_bonobo_set_sensitive (shell->priv->ui_component,
 					CMD_PATH_VIEW_SOURCELIST, TRUE);
   
 		gtk_widget_show (GTK_WIDGET (shell->priv->paned));
+ 		rb_statusbar_sync_state (shell->priv->statusbar);
+ 		gtk_widget_show (GTK_WIDGET (shell->priv->hsep));	
 	}
 
 	rb_source_header_sync_control_state (shell->priv->source_header);
@@ -1846,6 +1982,59 @@ rb_shell_cmd_current_song (BonoboUIComponent *component,
 	rb_shell_jump_to_current (shell);
 }
 
+static void
+rb_shell_jump_to_entry_with_source (RBShell *shell, RBSource *source,
+				    RhythmDBEntry *entry)
+{
+	RBEntryView *songs;
+
+	g_return_if_fail (entry != NULL);
+
+	if (source == NULL) {
+		const char *location;
+		rhythmdb_read_lock (shell->priv->db);
+		location = rhythmdb_entry_get_string (shell->priv->db, entry,
+						      RHYTHMDB_PROP_LOCATION);
+		rhythmdb_read_unlock (shell->priv->db);
+		if (rb_uri_is_iradio (location))
+			source = RB_SOURCE (shell->priv->iradio_source);
+		else
+			source = RB_SOURCE (shell->priv->library_source);
+	}
+
+	songs = rb_source_get_entry_view (source);
+	if (!rb_entry_view_get_entry_visible (songs, entry)) {
+		rb_source_reset_filters (source);
+		rb_source_header_clear_search (shell->priv->source_header);
+	}
+
+	rb_shell_select_source (shell, source);
+
+	while (gtk_events_pending ())
+		gtk_main_iteration ();
+
+	if (!rb_entry_view_get_entry_visible (songs, entry)) {
+		rb_source_search (shell->priv->selected_source, NULL);
+	}
+
+	rb_entry_view_scroll_to_entry (songs, entry);
+	rb_entry_view_select_entry (songs, entry);
+}
+
+static void
+rb_shell_jump_to_entry (RBShell *shell, RhythmDBEntry *entry)
+{
+	rb_shell_jump_to_entry_with_source (shell, NULL, entry);
+}
+
+static void
+rb_shell_play_entry (RBShell *shell, RhythmDBEntry *entry)
+{
+	rb_shell_player_stop (shell->priv->player_shell);
+	rb_shell_jump_to_entry_with_source (shell, NULL, entry);
+	rb_shell_player_play_entry (shell->priv->player_shell, entry);
+}
+
 static void
 rb_shell_jump_to_current (RBShell *shell)
 {
@@ -1856,20 +2045,9 @@ rb_shell_jump_to_current (RBShell *shell)
 	g_return_if_fail (source != NULL);
 
 	songs = rb_source_get_entry_view (source);
-
 	playing = rb_entry_view_get_playing_entry (songs);
 
-	g_return_if_fail (playing != NULL);
-
-	if (!rb_entry_view_get_entry_visible (songs, playing)) {
-		rb_source_reset_filters (source);
-		rb_source_header_clear_search (shell->priv->source_header);
-	}
-	rb_shell_select_source (shell, source);
-	if (!rb_entry_view_get_entry_visible (songs, playing)) {
-		rb_source_search (shell->priv->selected_source, NULL);
-	}
-	rb_shell_player_jump_to_current (shell->priv->player_shell);
+	rb_shell_jump_to_entry_with_source (shell, source, playing);
 }
 
 static gboolean
