@@ -41,7 +41,6 @@
 #include "rb-stock-icons.h"
 #include "rb-bonobo-helpers.h"
 #include "rb-debug.h"
-#include "rb-library-dnd-types.h"
 #include "eel-gconf-extensions.h"
 
 static void rb_tray_icon_class_init (RBTrayIconClass *klass);
@@ -88,7 +87,7 @@ struct RBTrayIconPrivate
 	GtkWindow *main_window;
 	GtkWidget *ebox;
 
-	RBLibrary *library;
+	RhythmDB *db;
 };
 
 enum
@@ -96,7 +95,7 @@ enum
 	PROP_0,
 	PROP_CONTAINER,
 	PROP_COMPONENT,
-	PROP_LIBRARY,
+	PROP_DB,
 	PROP_TRAY_COMPONENT,
 	PROP_WINDOW,
 };
@@ -120,10 +119,7 @@ static RBBonoboUIListener rb_tray_icon_listeners[] =
 	RB_BONOBO_UI_LISTENER_END
 };
 
-static const GtkTargetEntry target_uri [] =
-{
-	{ RB_LIBRARY_DND_URI_LIST_TYPE, 0, RB_LIBRARY_DND_URI_LIST }
-};
+static const GtkTargetEntry target_uri [] = {{ "text/uri-list", 0, 0 }};
 
 static GObjectClass *parent_class = NULL;
 
@@ -191,11 +187,11 @@ rb_tray_icon_class_init (RBTrayIconClass *klass)
 							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
 	g_object_class_install_property (object_class,
-					 PROP_LIBRARY,
-					 g_param_spec_object ("library",
-							      "RBLibrary",
-							      "RBLibrary object",
-							      RB_TYPE_LIBRARY,
+					 PROP_DB,
+					 g_param_spec_object ("db",
+							      "RhythmDB",
+							      "RhythmDB object",
+							      RHYTHMDB_TYPE,
 							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
 	g_object_class_install_property (object_class,
@@ -286,8 +282,8 @@ rb_tray_icon_set_property (GObject *object,
 	case PROP_COMPONENT:
 		tray->priv->main_component = g_value_get_object (value);
 		break;
-	case PROP_LIBRARY:
-		tray->priv->library = g_value_get_object (value);
+	case PROP_DB:
+		tray->priv->db = g_value_get_object (value);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -317,8 +313,8 @@ rb_tray_icon_get_property (GObject *object,
 	case PROP_TRAY_COMPONENT:
 		g_value_set_object (value, tray->priv->tray_component);
 		break;
-	case PROP_LIBRARY:
-		g_value_set_object (value, tray->priv->library);
+	case PROP_DB:
+		g_value_set_object (value, tray->priv->db);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -329,14 +325,14 @@ rb_tray_icon_get_property (GObject *object,
 RBTrayIcon *
 rb_tray_icon_new (BonoboUIContainer *container,
 		  BonoboUIComponent *component,
-		  RBLibrary *library,
+		  RhythmDB *db,
 		  GtkWindow *window)
 {
 	RBTrayIcon *tray = g_object_new (RB_TYPE_TRAY_ICON,
 					 "title", "Rhythmbox tray icon",
 					 "container", container,
 					 "component", component,
-					 "library", library,
+					 "db", db,
 					 "window", window,
 					 NULL);
 	
@@ -418,7 +414,7 @@ rb_tray_icon_drop_cb (GtkWidget *widget,
 		char *uri = i->data;
 
 		if (uri != NULL)
-			rb_library_add_uri (icon->priv->library, uri);
+			rhythmdb_add_uri_async (icon->priv->db, uri);
 
 		g_free (uri);
 	}
