@@ -1,7 +1,6 @@
 /* 
  *  arch-tag: Implementation of Internet Radio source object
  *
- *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
  *  Copyright (C) 2002,2003 Colin Walters <walters@debian.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -25,18 +24,24 @@
 #include <glade/glade.h>
 #include <libgnome/gnome-i18n.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
+#include <libxml/tree.h>
 #include <bonobo/bonobo-ui-component.h>
 #include <bonobo/bonobo-window.h>
 #include <string.h>
 
 #include "rb-iradio-source.h"
 
+#include "rhythmdb-legacy.h"
+#include "rhythmdb-model.h"
+#include "rhythmdb-query-model.h"
 #include "rb-bonobo-helpers.h"
 #include "rb-glade-helpers.h"
 #include "rb-stock-icons.h"
-#include "rb-node-view.h"
+#include "rb-entry-view.h"
+#include "rb-property-view.h"
 #include "rb-util.h"
 #include "rb-file-helpers.h"
+#include "rb-playlist.h"
 #include "rb-preferences.h"
 #include "rb-dialog.h"
 #include "rb-new-station-dialog.h"
@@ -44,10 +49,18 @@
 #include "rb-volume.h"
 #include "rb-debug.h"
 #include "eel-gconf-extensions.h"
-#include "rb-node-filter.h"
+
+typedef enum
+{
+	RB_IRADIO_QUERY_TYPE_ALL,
+	RB_IRADIO_QUERY_TYPE_GENRE,
+	RB_IRADIO_QUERY_TYPE_SEARCH,
+} RBIRadioQueryType;
 
 static void rb_iradio_source_class_init (RBIRadioSourceClass *klass);
 static void rb_iradio_source_init (RBIRadioSource *source);
+static GObject *rb_iradio_source_constructor (GType type, guint n_construct_properties,
+					      GObjectConstructParam *construct_properties);
 static void rb_iradio_source_finalize (GObject *object);
 static void rb_iradio_source_set_property (GObject *object,
 			                  guint prop_id,
@@ -57,12 +70,9 @@ static void rb_iradio_source_get_property (GObject *object,
 			                  guint prop_id,
 			                  GValue *value,
 			                  GParamSpec *pspec);
-static void genre_node_selected_cb (RBNodeView *view,
-				     RBNode *node,
-				     RBIRadioSource *source);
-static void rb_iradio_source_songs_show_popup_cb (RBNodeView *view,
-						RBNode *node,
-						RBIRadioSource *source);
+static void rb_iradio_source_songs_show_popup_cb (RBEntryView *view,
+						  RhythmDBEntry *entry,
+						  RBIRadioSource *source);
 static void paned_size_allocate_cb (GtkWidget *widget,
 				    GtkAllocation *allocation,
 		                    RBIRadioSource *source);
@@ -70,25 +80,29 @@ static void rb_iradio_source_state_pref_changed (GConfClient *client,
 						 guint cnxn_id,
 						 GConfEntry *entry,
 						 RBIRadioSource *source);
-
+static void rb_iradio_source_first_time_changed (GConfClient *client,
+						 guint cnxn_id,
+						 GConfEntry *entry,
+						 RBIRadioSource *source);
 static void rb_iradio_source_show_browser (RBIRadioSource *source,
 					   gboolean show);
-
 static void rb_iradio_source_state_prefs_sync (RBIRadioSource *source);
+static void genre_selected_cb (RBPropertyView *propview, const char *name,
+			       RBIRadioSource *iradio_source);
+static void rb_iradio_source_songs_view_sort_order_changed_cb (RBEntryView *view, RBIRadioSource *source);
 
 /* source methods */
 static const char *impl_get_status (RBSource *source);
 static const char *impl_get_browser_key (RBSource *source);
 static GdkPixbuf *impl_get_pixbuf (RBSource *source);
-static RBNodeView *impl_get_node_view (RBSource *source);
+static RBEntryView *impl_get_entry_view (RBSource *source);
 static void impl_search (RBSource *source, const char *text);
 static void impl_delete (RBSource *source);
 static void impl_song_properties (RBSource *source);
 static RBSourceEOFType impl_handle_eos (RBSource *asource);
 static void impl_buffering_done (RBSource *asource);
+static void rb_iradio_source_do_query (RBIRadioSource *source, RBIRadioQueryType type);
 
-static void stations_filter (RBIRadioSource *source,
-			     RBNode *genre);
 void rb_iradio_source_show_columns_changed_cb (GtkToggleButton *button,
 					     RBIRadioSource *source);
 
@@ -101,18 +115,19 @@ void rb_iradio_source_show_columns_changed_cb (GtkToggleButton *button,
 #define CONF_UI_IRADIO_COLUMNS_SETUP CONF_PREFIX "/ui/iradio/columns_setup"
 #define CONF_STATE_IRADIO_DIR CONF_PREFIX "/state/iradio"
 #define CONF_STATE_PANED_POSITION CONF_PREFIX "/state/iradio/paned_position"
+#define CONF_STATE_IRADIO_SORTING CONF_PREFIX "/state/iradio/sorting"
 #define CONF_STATE_SHOW_BROWSER   CONF_PREFIX "/state/iradio/show_browser"
 
 struct RBIRadioSourcePrivate
 {
-	RBIRadioBackend *backend;
+	RhythmDB *db;
 
 	GtkWidget *vbox;
 
 	GdkPixbuf *pixbuf;
 
-	RBNodeView *genres;
-	RBNodeView *stations;
+	RBPropertyView *genres;
+	RBEntryView *stations;
 
 	gboolean async_playcount_update_queued;
 	guint async_playcount_update_id;
@@ -129,11 +144,12 @@ struct RBIRadioSourcePrivate
 
 	gboolean lock;
 
-	gboolean changing_genre;
-	RBNodeFilter *stations_filter;
-
 	gboolean loading_prefs;
 
+	RBIRadioQueryType query_type;
+	char *search_text;
+	char *selected_genre;
+
 	GtkWidget *genre_check;
 	GtkWidget *rating_check;
 	GtkWidget *play_count_check;
@@ -141,15 +157,15 @@ struct RBIRadioSourcePrivate
 	GtkWidget *quality_check;
 
 	guint async_idlenum;
-	gboolean async_node_destroyed;
+	gboolean async_entry_destroyed;
 	guint async_signum;
-	RBNode *async_update_node;
+	RhythmDB *async_update_entry;
 };
 
 enum
 {
 	PROP_0,
-	PROP_BACKEND
+	PROP_DB,
 };
 
 static GObjectClass *parent_class = NULL;
@@ -192,6 +208,7 @@ rb_iradio_source_class_init (RBIRadioSourceClass *klass)
 	parent_class = g_type_class_peek_parent (klass);
 
 	object_class->finalize = rb_iradio_source_finalize;
+	object_class->constructor = rb_iradio_source_constructor;
 
 	object_class->set_property = rb_iradio_source_set_property;
 	object_class->get_property = rb_iradio_source_get_property;
@@ -201,7 +218,7 @@ rb_iradio_source_class_init (RBIRadioSourceClass *klass)
 	source_class->impl_get_pixbuf  = impl_get_pixbuf;
 	source_class->impl_can_search = (RBSourceFeatureFunc) rb_true_function;
 	source_class->impl_search = impl_search;
-	source_class->impl_get_node_view = impl_get_node_view;
+	source_class->impl_get_entry_view = impl_get_entry_view;
 	source_class->impl_can_delete = (RBSourceFeatureFunc) rb_true_function;
 	source_class->impl_delete = impl_delete;
 	source_class->impl_song_properties = impl_song_properties;
@@ -212,11 +229,11 @@ rb_iradio_source_class_init (RBIRadioSourceClass *klass)
 	source_class->impl_buffering_done = impl_buffering_done;
 
 	g_object_class_install_property (object_class,
-					 PROP_BACKEND,
-					 g_param_spec_object ("backend",
-							      "Backend",
-							      "IRadio Backend",
-							      RB_TYPE_IRADIO_BACKEND,
+					 PROP_DB,
+					 g_param_spec_object ("db",
+							      "RhythmDB",
+							      "RhythmDB database",
+							      RHYTHMDB_TYPE,
 							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	
 }
@@ -235,11 +252,9 @@ rb_iradio_source_init (RBIRadioSource *source)
 	source->priv->name = g_strdup (_("Unknown"));
 	source->priv->title = g_strdup (_("Unknown"));
 
-	source->priv->changing_genre = FALSE;
-
 	source->priv->pixbuf = gtk_widget_render_icon (dummy,
 						       RB_STOCK_IRADIO,
-						       GTK_ICON_SIZE_MENU,
+						       GTK_ICON_SIZE_LARGE_TOOLBAR,
 						       NULL);
 	gtk_widget_destroy (dummy);
 }
@@ -263,13 +278,78 @@ rb_iradio_source_finalize (GObject *object)
 	g_free (source->priv->title);
 	g_free (source->priv->status);
 
-	g_object_unref (G_OBJECT (source->priv->stations_filter));
-
 	g_free (source->priv);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static GObject *
+rb_iradio_source_constructor (GType type, guint n_construct_properties,
+			      GObjectConstructParam *construct_properties)
+{
+	RBIRadioSource *source;
+	RBIRadioSourceClass *klass;
+	GObjectClass *parent_class;  
+	klass = RB_IRADIO_SOURCE_CLASS (g_type_class_peek (type));
+
+	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
+	source = RB_IRADIO_SOURCE (parent_class->constructor (type, n_construct_properties,
+							      construct_properties));
+
+	source->priv->paned = gtk_hpaned_new ();
+
+	/* set up stations view */
+	source->priv->stations = rb_entry_view_new (source->priv->db, CONF_STATE_IRADIO_SORTING,
+						    FALSE, FALSE);
+	rb_entry_view_append_column (source->priv->stations, RB_ENTRY_VIEW_COL_TITLE);
+	rb_entry_view_append_column (source->priv->stations, RB_ENTRY_VIEW_COL_GENRE);
+	rb_entry_view_append_column (source->priv->stations, RB_ENTRY_VIEW_COL_RATING);
+	rb_entry_view_append_column (source->priv->stations, RB_ENTRY_VIEW_COL_PLAY_COUNT);
+	rb_entry_view_append_column (source->priv->stations, RB_ENTRY_VIEW_COL_LAST_PLAYED);
+	g_signal_connect (G_OBJECT (source->priv->stations),
+			  "sort-order-changed",
+			  G_CALLBACK (rb_iradio_source_songs_view_sort_order_changed_cb),
+			  source);
+
+	g_signal_connect (G_OBJECT (source->priv->stations),
+			  "size_allocate",
+			  G_CALLBACK (paned_size_allocate_cb),
+			  source);
+	g_signal_connect (G_OBJECT (source->priv->stations), "show_popup",
+			  G_CALLBACK (rb_iradio_source_songs_show_popup_cb), source);
+
+	/* set up genre entry view */
+	source->priv->genres = rb_property_view_new (source->priv->db,
+						     RHYTHMDB_PROP_GENRE,
+						     _("Genre"));
+	g_signal_connect (G_OBJECT (source->priv->genres),
+			  "property-selected",
+			  G_CALLBACK (genre_selected_cb),
+			  source);
+
+	g_object_set (G_OBJECT (source->priv->genres), "vscrollbar_policy",
+		      GTK_POLICY_AUTOMATIC, NULL);
+	g_object_ref (G_OBJECT (source->priv->genres));
+
+
+	gtk_paned_pack2 (GTK_PANED (source->priv->paned),
+			 GTK_WIDGET (source->priv->stations), TRUE, FALSE);
+
+	gtk_box_pack_start_defaults (GTK_BOX (source->priv->vbox), source->priv->paned);
+
+	rb_iradio_source_state_prefs_sync (source);
+	eel_gconf_notification_add (CONF_STATE_IRADIO_DIR,
+				    (GConfClientNotifyFunc) rb_iradio_source_state_pref_changed,
+				    source);
+	eel_gconf_notification_add (CONF_FIRST_TIME,
+				    (GConfClientNotifyFunc) rb_iradio_source_first_time_changed,
+				    source);
+	rb_iradio_source_do_query (source, RB_IRADIO_QUERY_TYPE_ALL);
+			
+	gtk_widget_show_all (GTK_WIDGET (source));
+	return G_OBJECT (source);
+}
+
 static void
 rb_iradio_source_set_property (GObject *object,
 			      guint prop_id,
@@ -280,51 +360,9 @@ rb_iradio_source_set_property (GObject *object,
 
 	switch (prop_id)
 	{
-	case PROP_BACKEND:
-	{
-		source->priv->backend = g_value_get_object (value);
-
-		source->priv->paned = gtk_hpaned_new ();
-
-		/* Initialize the filters */
-		source->priv->stations_filter = rb_node_filter_new ();
-
-		/* set up genre node view */
-		source->priv->genres = rb_node_view_new (rb_iradio_backend_get_all_genres (source->priv->backend),
-						       rb_file ("rb-node-view-iradio-genres.xml"),
-						       NULL);
-
-		g_object_set (G_OBJECT (source->priv->genres), "vscrollbar_policy", GTK_POLICY_AUTOMATIC, NULL);
-		g_object_ref (G_OBJECT (source->priv->genres));
-
-		g_signal_connect (G_OBJECT (source->priv->genres),
-				  "node_selected",
-				  G_CALLBACK (genre_node_selected_cb),
-				  source);
-
-		/* set up stations tree view */
-		source->priv->stations = rb_node_view_new (rb_iradio_backend_get_all_stations (source->priv->backend),
-							 rb_file ("rb-node-view-iradio-stations.xml"),
-							 source->priv->stations_filter);
-		g_signal_connect (G_OBJECT (source->priv->stations),
-				  "size_allocate",
-				  G_CALLBACK (paned_size_allocate_cb),
-				  source);
-		g_signal_connect (G_OBJECT (source->priv->stations), "show_popup",
-				  G_CALLBACK (rb_iradio_source_songs_show_popup_cb), source);
-
-		gtk_paned_pack2 (GTK_PANED (source->priv->paned), GTK_WIDGET (source->priv->stations), TRUE, FALSE);
-
-		gtk_box_pack_start_defaults (GTK_BOX (source->priv->vbox), source->priv->paned);
-
-		rb_iradio_source_state_prefs_sync (source);
-		eel_gconf_notification_add (CONF_STATE_IRADIO_DIR,
-					    (GConfClientNotifyFunc) rb_iradio_source_state_pref_changed,
-					    source);
-			
-		gtk_widget_show_all (GTK_WIDGET (source));
+	case PROP_DB:
+		source->priv->db = g_value_get_object (value);
 		break;
-	}
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -341,8 +379,8 @@ rb_iradio_source_get_property (GObject *object,
 
 	switch (prop_id)
 	{
-	case PROP_BACKEND:
-		g_value_set_object (value, source->priv->backend);
+	case PROP_DB:
+		g_value_set_object (value, source->priv->db);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -351,18 +389,43 @@ rb_iradio_source_get_property (GObject *object,
 }
 
 RBSource *
-rb_iradio_source_new (RBIRadioBackend *backend)
+rb_iradio_source_new (RhythmDB *db)
 {
 	RBSource *source;
 
 	source = RB_SOURCE (g_object_new (RB_TYPE_IRADIO_SOURCE,
 					  "name", _("Radio"),
-					  "backend", backend,
+					  "db", db,
 					  NULL));
 
 	return source;
 }
 
+void
+rb_iradio_source_add_station (RBIRadioSource *source,
+			      const char *uri, const char *title, const char *genre)
+{
+	RhythmDBEntry *entry;
+	GValue val = { 0, };
+
+	rhythmdb_write_lock (source->priv->db);
+
+	entry = rhythmdb_entry_lookup_by_location (source->priv->db, uri);
+	if (!entry)
+		entry = rhythmdb_entry_new (source->priv->db, RHYTHMDB_ENTRY_TYPE_IRADIO_STATION, uri);
+	g_value_init (&val, G_TYPE_STRING);
+	g_value_set_string (&val, title);
+	rhythmdb_entry_set (source->priv->db, entry, RHYTHMDB_PROP_TITLE, &val);
+	if (genre) {
+		g_value_reset (&val);
+		g_value_set_string (&val, genre);
+		rhythmdb_entry_set (source->priv->db, entry, RHYTHMDB_PROP_GENRE, &val);
+	}
+	g_value_unset (&val);
+		
+	rhythmdb_write_unlock (source->priv->db);
+}
+
 static GdkPixbuf *
 impl_get_pixbuf (RBSource *asource)
 {
@@ -375,27 +438,25 @@ static void
 impl_search (RBSource *asource, const char *search_text)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
+	if (search_text == NULL && source->priv->search_text == NULL)
+		return;
+	if (search_text != NULL &&
+	    source->priv->search_text != NULL
+	    && !strcmp (search_text, source->priv->search_text))
+		return;
+	if (search_text[0] == '\0')
+	  search_text = NULL;
 
-	/* resets the filter */
-	if (search_text == NULL || strcmp (search_text, "") == 0) {
-		rb_node_view_select_node (source->priv->genres,
-		 		          rb_iradio_backend_get_all_stations (source->priv->backend));
-	} else {
-		rb_node_view_select_none (source->priv->genres);
-
-		rb_node_filter_empty (source->priv->stations_filter);
-		rb_node_filter_add_expression (source->priv->stations_filter,
-					       rb_node_filter_expression_new (source->priv->stations_filter,
-									      RB_NODE_FILTER_EXPRESSION_STRING_PROP_CONTAINS,
-									      RB_NODE_PROP_NAME,
-									      search_text),
-					       0);
-		rb_node_filter_done_changing (source->priv->stations_filter);
-	}
+	g_free (source->priv->search_text);
+	if (search_text)
+		source->priv->search_text = g_utf8_casefold (search_text, -1);
+	else
+		source->priv->search_text = NULL;
+	rb_iradio_source_do_query (source, RB_IRADIO_QUERY_TYPE_SEARCH);
 }
 
-static RBNodeView *
-impl_get_node_view (RBSource *asource)
+static RBEntryView *
+impl_get_entry_view (RBSource *asource)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
 
@@ -411,44 +472,31 @@ impl_handle_eos (RBSource *asource)
 struct RBIRadioAsyncPlayStatisticsData
 {
 	RBIRadioSource *source;
-	RBNode *node;
+	RhythmDBEntry *entry;
 };
 
-static void
-async_node_update_destroyed_cb (RBNode *node, gpointer data)
-{
-	RBIRadioSource *source = RB_IRADIO_SOURCE (data);
-	rb_debug ("node to be updated was destroyed");
-
-	source->priv->async_node_destroyed = TRUE;
-}
-
 static gboolean
 rb_iradio_source_async_update_play_statistics (gpointer data)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (data);
-	RBNode *playing_node;
+	RhythmDBEntry *playing_entry;
 
 	rb_debug ("entering async handler");
 
-	gdk_threads_enter ();
+	GDK_THREADS_ENTER ();
 
-
-	if (!source->priv->async_node_destroyed) {
-		playing_node = rb_node_view_get_playing_node (source->priv->stations);
-		rb_debug ("async updating play statistics, node: %p playing node: %p",
-			  source->priv->async_update_node, playing_node);
-		if (source->priv->async_update_node == playing_node)
-			rb_node_update_play_statistics (source->priv->async_update_node);
-
-		rb_node_signal_disconnect (source->priv->async_update_node,
-					   source->priv->async_signum);
-	} else {
-		rb_debug ("async node destroyed");
-	}
-	source->priv->async_update_node = NULL;
+	playing_entry = rb_entry_view_get_playing_entry (source->priv->stations);
+	rb_debug ("async updating play statistics, entry: %p playing entry: %p",
+		  source->priv->async_update_entry, playing_entry);
+	if (source->priv->async_update_entry == playing_entry)
+		rb_source_update_play_statistics (RB_SOURCE (source),
+						  source->priv->db,
+						  source->priv->async_update_entry);
+	rhythmdb_entry_unref (source->priv->db, source->priv->async_update_entry);
+	source->priv->async_update_entry = NULL;
 		
-	gdk_threads_leave ();
+	GDK_THREADS_LEAVE ();
+
 	return FALSE;
 }
 
@@ -456,23 +504,19 @@ static void
 impl_buffering_done (RBSource *asource)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
-	RBNode *node = rb_node_view_get_playing_node (source->priv->stations);
+	RhythmDBEntry *entry = rb_entry_view_get_playing_entry (source->priv->stations);
 
-	rb_debug ("queueing async play statistics update, node: %p", node);
+	rb_debug ("queueing async play statistics update, entry: %p", entry);
 
-	if (source->priv->async_update_node != NULL) {
+	if (source->priv->async_update_entry != NULL) {
 		rb_debug ("async handler already queued, removing");
-		rb_node_signal_disconnect (source->priv->async_update_node,
-					   source->priv->async_signum);
+		rhythmdb_entry_unref (source->priv->db, source->priv->async_update_entry);
 		g_source_remove (source->priv->async_idlenum);
 	}
 
-	source->priv->async_node_destroyed = FALSE;
-	source->priv->async_update_node = node;
- 	source->priv->async_signum =
-		rb_node_signal_connect_object (node, RB_NODE_DESTROY,
-					       (RBNodeCallback) async_node_update_destroyed_cb,
-					       G_OBJECT (source));
+	source->priv->async_entry_destroyed = FALSE;
+	source->priv->async_update_entry = entry;
+	rhythmdb_entry_ref (source->priv->db, source->priv->async_update_entry);
 	source->priv->async_idlenum =
 		g_timeout_add (6000, (GSourceFunc) rb_iradio_source_async_update_play_statistics,
 			       source);
@@ -483,9 +527,9 @@ impl_get_status (RBSource *asource)
 {
  	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
 	char *ret;
-	ret = g_strdup_printf (_("<b>%d</b> total stations in <b>%d</b> distinct genres"),
-			       rb_iradio_backend_get_station_count (source->priv->backend),
-			       rb_iradio_backend_get_genre_count (source->priv->backend));
+	guint num_entries = rb_entry_view_get_num_entries (source->priv->stations);
+	ret = g_strdup_printf (ngettext ("%d station", "%d stations", num_entries),
+			       num_entries);
 	return ret;
 }
 
@@ -501,16 +545,19 @@ impl_delete (RBSource *asource)
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
 	GList *l;
 
-	for (l = rb_node_view_get_selection (source->priv->stations); l != NULL; l = g_list_next (l))
-		rb_iradio_backend_remove_node (source->priv->backend, l->data);
+	for (l = rb_entry_view_get_selected_entries (source->priv->stations); l != NULL;
+	     l = g_list_next (l)) {
+		rhythmdb_write_lock (source->priv->db);
+		rhythmdb_entry_delete (source->priv->db, l->data);
+		rhythmdb_write_unlock (source->priv->db);
+	}
 }
 
 static void
 impl_song_properties (RBSource *asource)
 {
 	RBIRadioSource *source = RB_IRADIO_SOURCE (asource);
-	GtkWidget *dialog = rb_station_properties_dialog_new (source->priv->stations,
-							      source->priv->backend);
+	GtkWidget *dialog = rb_station_properties_dialog_new (source->priv->stations);
 	rb_debug ("in song properties");
 	if (dialog)
 		gtk_widget_show_all (dialog);
@@ -549,10 +596,17 @@ rb_iradio_source_state_pref_changed (GConfClient *client,
 	rb_iradio_source_state_prefs_sync (source);
 }
 
+static void
+rb_iradio_source_songs_view_sort_order_changed_cb (RBEntryView *view,
+						   RBIRadioSource *source)
+{
+	rb_debug ("sort order changed");
+	rb_iradio_source_do_query (source, RB_IRADIO_QUERY_TYPE_SEARCH);
+}
 
 static void
-rb_iradio_source_songs_show_popup_cb (RBNodeView *view,
-				      RBNode *node,
+rb_iradio_source_songs_show_popup_cb (RBEntryView *view,
+				      RhythmDBEntry *entry,
 				      RBIRadioSource *source)
 {
 	GtkWidget *menu;
@@ -572,38 +626,13 @@ rb_iradio_source_songs_show_popup_cb (RBNodeView *view,
 	gtk_object_sink (GTK_OBJECT (menu));
 }
 
-static RBNode *
-ensure_node_selection (RBNodeView *view,
-		       RBNode *all_node,
-		       gboolean *changing_flag)
-{
-	GList *selection = rb_node_view_get_selection (view);
-
-	if (selection == NULL) {
-		*changing_flag = TRUE;
-		rb_node_view_select_node (view, all_node);
-		*changing_flag = FALSE;
-		selection = rb_node_view_get_selection (view);
-	}
-
-	return selection->data;
-}
-
 static void
-genre_node_selected_cb (RBNodeView *view,
-			RBNode *node,
-			RBIRadioSource *source)
+genre_selected_cb (RBPropertyView *propview, const char *name,
+		   RBIRadioSource *iradio_source)
 {
-	RBNode *genre = NULL;
-	if (source->priv->changing_genre == TRUE)
-		return;
-
-	genre = ensure_node_selection (view,
-				       rb_iradio_backend_get_all_genres(source->priv->backend),
-				       &source->priv->changing_genre);
-
-	rb_source_notify_filter_changed (RB_SOURCE (source));
-	stations_filter (source, genre);
+	g_free (iradio_source->priv->selected_genre);
+	iradio_source->priv->selected_genre = g_strdup (name);
+	rb_iradio_source_do_query (iradio_source, RB_IRADIO_QUERY_TYPE_GENRE);
 }
 
 static void
@@ -625,19 +654,101 @@ rb_iradio_source_show_browser (RBIRadioSource *source,
 }
 
 static void
-stations_filter (RBIRadioSource *source,
-		 RBNode *parent)
+rb_iradio_source_do_query (RBIRadioSource *source, RBIRadioQueryType qtype)
 {
-	rb_node_filter_empty (source->priv->stations_filter);
-	rb_node_filter_add_expression (source->priv->stations_filter,
-				       rb_node_filter_expression_new (source->priv->stations_filter,
-								      RB_NODE_FILTER_EXPRESSION_EQUALS,
-								      rb_iradio_backend_get_all_stations (source->priv->backend)),
-				       0);
-	rb_node_filter_add_expression (source->priv->stations_filter,
-				       rb_node_filter_expression_new (source->priv->stations_filter,
-								      RB_NODE_FILTER_EXPRESSION_HAS_PARENT,
-								      parent),
-				       0);
-	rb_node_filter_done_changing (source->priv->stations_filter);
+	RhythmDBQueryModel *query_model;
+	RhythmDBPropertyModel *genre_model;
+	GtkTreeModel *model;
+	GPtrArray *query;
+
+	rhythmdb_read_lock (source->priv->db);
+	source->priv->query_type = qtype;
+	query = rhythmdb_query_parse (source->priv->db,
+				      RHYTHMDB_QUERY_PROP_EQUALS,
+				      RHYTHMDB_PROP_TYPE,
+				      RHYTHMDB_ENTRY_TYPE_IRADIO_STATION,
+				      RHYTHMDB_QUERY_END);
+
+	if (source->priv->search_text) {
+		GPtrArray *subquery = rhythmdb_query_parse (source->priv->db,
+							    RHYTHMDB_QUERY_PROP_LIKE,
+							    RHYTHMDB_PROP_GENRE_FOLDED,
+							    source->priv->search_text,
+							    RHYTHMDB_QUERY_DISJUNCTION,
+							    RHYTHMDB_QUERY_PROP_LIKE,
+							    RHYTHMDB_PROP_TITLE_FOLDED,
+							    source->priv->search_text,
+							    RHYTHMDB_QUERY_END);
+		rhythmdb_query_append (source->priv->db,
+				       query,
+				       RHYTHMDB_QUERY_SUBQUERY,
+				       subquery,
+				       RHYTHMDB_QUERY_END);
+	}
+
+	query_model = rhythmdb_query_model_new_empty (source->priv->db);
+
+	if (qtype < RB_IRADIO_QUERY_TYPE_GENRE) {
+		rb_property_view_reset (source->priv->genres);
+		g_free (source->priv->selected_genre);
+		source->priv->selected_genre = NULL;
+	} 
+
+	genre_model = rb_property_view_get_model (source->priv->genres);
+
+	if (qtype < RB_IRADIO_QUERY_TYPE_GENRE) {
+		g_object_set (G_OBJECT (genre_model), "query-model", query_model, NULL);
+	} else
+		g_object_set (G_OBJECT (genre_model), "query-model", NULL, NULL);
+
+	if (source->priv->selected_genre)
+		rhythmdb_query_append (source->priv->db,
+				       query,
+				       RHYTHMDB_QUERY_PROP_EQUALS,
+				       RHYTHMDB_PROP_GENRE,
+				       source->priv->selected_genre,
+				       RHYTHMDB_QUERY_END);
+
+	model = GTK_TREE_MODEL (query_model);
+	
+	rb_entry_view_set_model (source->priv->stations, RHYTHMDB_MODEL (query_model));
+
+	rhythmdb_do_full_query_parsed (source->priv->db, model, query);
+
+	if (qtype >= RB_IRADIO_QUERY_TYPE_GENRE)
+		g_object_set (G_OBJECT (genre_model), "query-model", query_model, NULL);
+
+	rhythmdb_query_free (query);
+
+	rhythmdb_read_unlock (source->priv->db);
 }
+
+static void
+handle_playlist_entry_cb (RBPlaylist *playlist, const char *uri, const char *title,
+			  const char *genre, RBIRadioSource *source)
+{
+	if (rb_uri_is_iradio (uri)) {
+		rhythmdb_read_lock (source->priv->db);
+		if (!rhythmdb_entry_lookup_by_location (source->priv->db, uri)) {
+			rhythmdb_read_unlock (source->priv->db);
+			rb_iradio_source_add_station (source, uri, title, genre);
+		} else
+			rhythmdb_read_unlock (source->priv->db);
+	}
+}
+
+static void
+rb_iradio_source_first_time_changed (GConfClient *client,
+				     guint cnxn_id,
+				     GConfEntry *entry,
+				     RBIRadioSource *source)
+{
+	RBPlaylist *parser = rb_playlist_new ();
+
+	g_signal_connect_object (G_OBJECT (parser), "entry",
+				 G_CALLBACK (handle_playlist_entry_cb),
+				 source, 0);
+	rb_playlist_parse (parser, rb_file ("iradio-initial.pls"));	
+	g_object_unref (G_OBJECT (parser));
+}
+
