@@ -61,12 +61,6 @@ typedef enum
 	RB_LIBRARY_QUERY_TYPE_SEARCH,
 } RBLibraryQueryType;
 
-struct RBLibrarySourceEntryAddData
-{
-	RBLibrarySource *source;
-	RBPropertyView *propview;
-};
-
 static void rb_library_source_class_init (RBLibrarySourceClass *klass);
 static void rb_library_source_init (RBLibrarySource *source);
 static GObject *rb_library_source_constructor (GType type, guint n_construct_properties,
@@ -89,10 +83,6 @@ static void artists_selection_reset_cb (RBPropertyView *propview, RBLibrarySourc
 static void albums_selected_cb (RBPropertyView *propview, GList *albums,
 			       RBLibrarySource *libsource);
 static void albums_selection_reset_cb (RBPropertyView *propview, RBLibrarySource *libsource);
-static void entry_added_cb (RBEntryView *view, RhythmDBEntry *entry,
-			    struct RBLibrarySourceEntryAddData *data);
-static void entry_deleted_cb (RBEntryView *view, RhythmDBEntry *entry,
-			      struct RBLibrarySourceEntryAddData *data);
 static void songs_view_sort_order_changed_cb (RBEntryView *view, RBLibrarySource *source);
 
 static void paned_size_allocate_cb (GtkWidget *widget,
@@ -170,14 +160,16 @@ struct RBLibrarySourcePrivate
 
 	gboolean lock;
 
-	guint genre_add_handler_id;
-	guint artist_add_handler_id;
-	guint album_add_handler_id;
-
 	char *status;
 	char *artist;
 	char *album;
 
+	RhythmDBQueryModel *cached_all_query;
+	RhythmDBPropertyModel *cached_genres_model;
+	RhythmDBPropertyModel *cached_artists_model;
+	RhythmDBPropertyModel *cached_albums_model;
+	char *cached_sorting_type;
+	
 	RhythmDBQueryModel *model;
 	RhythmDBQueryModel *active_query;
 	RBLibraryQueryType query_type;
@@ -366,6 +358,10 @@ rb_library_source_finalize (GObject *object)
 	g_free (source->priv->artist);
 	g_free (source->priv->album);
 
+	g_free (source->priv->cached_sorting_type);
+	if (source->priv->cached_all_query)
+		g_object_unref (G_OBJECT (source->priv->cached_all_query));
+
 	g_free (source->priv);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -399,7 +395,6 @@ rb_library_source_constructor (GType type, guint n_construct_properties,
 	RBLibrarySource *source;
 	RBLibrarySourceClass *klass;
 	GObjectClass *parent_class;  
-	struct RBLibrarySourceEntryAddData *add_data;
 
 	klass = RB_LIBRARY_SOURCE_CLASS (g_type_class_peek (type));
 
@@ -439,16 +434,6 @@ rb_library_source_constructor (GType type, guint n_construct_properties,
 			  "properties-selected",
 			  G_CALLBACK (genres_selected_cb),
 			  source);
-	add_data = g_new0 (struct RBLibrarySourceEntryAddData, 1);
-	add_data->source = source;
-	add_data->propview = source->priv->genres;
-	source->priv->genre_add_handler_id =
-		g_signal_connect (G_OBJECT (source->priv->songs),
-				  "entry-added", G_CALLBACK (entry_added_cb),
-				  add_data);
-	g_signal_connect (G_OBJECT (source->priv->songs),
-			  "entry-deleted", G_CALLBACK (entry_deleted_cb),
-			  add_data);
 	g_signal_connect (G_OBJECT (source->priv->genres),
 			  "property-selection-reset", G_CALLBACK (genres_selection_reset_cb),
 			  source);
@@ -463,16 +448,6 @@ rb_library_source_constructor (GType type, guint n_construct_properties,
 			  "properties-selected",
 			  G_CALLBACK (artists_selected_cb),
 			  source);
-	add_data = g_new0 (struct RBLibrarySourceEntryAddData, 1);
-	add_data->source = source;
-	add_data->propview = source->priv->artists;
-	source->priv->artist_add_handler_id =
-		g_signal_connect (G_OBJECT (source->priv->songs),
-				  "entry-added", G_CALLBACK (entry_added_cb),
-				  add_data);
-	g_signal_connect (G_OBJECT (source->priv->songs),
-			  "entry-deleted", G_CALLBACK (entry_deleted_cb),
-			  add_data);
 	g_signal_connect (G_OBJECT (source->priv->artists),
 			  "property-selection-reset", G_CALLBACK (artists_selection_reset_cb),
 			  source);
@@ -487,16 +462,6 @@ rb_library_source_constructor (GType type, guint n_construct_properties,
 			  "properties-selected",
 			  G_CALLBACK (albums_selected_cb),
 			  source);
-	add_data = g_new0 (struct RBLibrarySourceEntryAddData, 1);
-	add_data->source = source;
-	add_data->propview = source->priv->albums;
-	source->priv->album_add_handler_id =
-		g_signal_connect (G_OBJECT (source->priv->songs),
-				  "entry-added", G_CALLBACK (entry_added_cb),
-				  add_data);
-	g_signal_connect (G_OBJECT (source->priv->songs),
-			  "entry-deleted", G_CALLBACK (entry_deleted_cb),
-			  add_data);
 	g_signal_connect (G_OBJECT (source->priv->albums),
 			  "property-selection-reset", G_CALLBACK (albums_selection_reset_cb),
 			  source);
@@ -1011,26 +976,38 @@ rb_library_source_state_pref_changed (GConfClient *client,
 	rb_library_source_state_prefs_sync (source);
 }
 
+static void
+rb_library_source_add_location_entry_changed_cb (GtkEntry *entry,
+						 GtkWidget *target)
+{
+	gtk_widget_set_sensitive (target, g_utf8_strlen (gtk_entry_get_text (GTK_ENTRY (entry)), -1) > 0);
+}
+
 void
 rb_library_source_add_location (RBLibrarySource *source, GtkWindow *win)
 {
 	GladeXML *xml = rb_glade_xml_new ("uri.glade",
 					  "open_uri_dialog_content",
 					  source);
-	GtkWidget *content, *uri_widget;
+	GtkWidget *content, *uri_widget, *open_button;
 	GtkWidget *dialog = gtk_dialog_new_with_buttons (_("Add Location"),
 							 win,
-							 GTK_DIALOG_MODAL,
+							 GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
 							 GTK_STOCK_CANCEL,
 							 GTK_RESPONSE_CANCEL,
-							 GTK_STOCK_OPEN,
-							 GTK_RESPONSE_OK,
 							 NULL);
 
 	g_return_if_fail (dialog != NULL);
 
+	open_button = gtk_dialog_add_button (GTK_DIALOG (dialog),
+					     GTK_STOCK_OPEN,
+					     GTK_RESPONSE_OK);
+	gtk_widget_set_sensitive (open_button, FALSE);
+
 	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
 					 GTK_RESPONSE_OK);
+	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
+	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
 
 	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
 
@@ -1040,11 +1017,17 @@ rb_library_source_add_location (RBLibrarySource *source, GtkWindow *win)
 
 	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
 			    content, FALSE, FALSE, 0);
+	gtk_container_set_border_width (GTK_CONTAINER (content), 5);
 
 	uri_widget = glade_xml_get_widget (xml, "uri");
 
 	g_return_if_fail (uri_widget != NULL);
 
+	g_signal_connect (G_OBJECT (uri_widget),
+			  "changed",
+			  G_CALLBACK (rb_library_source_add_location_entry_changed_cb),
+			  open_button);
+
 	switch (gtk_dialog_run (GTK_DIALOG (dialog))) {
 	case GTK_RESPONSE_OK:
 	{
@@ -1111,20 +1094,6 @@ impl_show_popup (RBSource *source)
 	return TRUE;
 }
 
-static void
-entry_added_cb (RBEntryView *view, RhythmDBEntry *entry,
-		struct RBLibrarySourceEntryAddData *data)
-{
-	rb_property_view_handle_entry_addition (data->propview, entry);
-}
-
-static void
-entry_deleted_cb (RBEntryView *view, RhythmDBEntry *entry,
-		  struct RBLibrarySourceEntryAddData *data)
-{
-	rb_property_view_handle_entry_deletion (data->propview, entry);
-}
-
 static void
 push_multi_equals_query (RhythmDB *db, GPtrArray *query, guint propid, GList *items)
 {
@@ -1249,122 +1218,178 @@ rb_library_source_do_query (RBLibrarySource *source, RBLibraryQueryType qtype,
 			    gboolean sync)
 {
 	RhythmDBQueryModel *query_model;
+	RhythmDBPropertyModel *genre_model = NULL;
+	RhythmDBPropertyModel *artist_model = NULL;
+	RhythmDBPropertyModel *album_model = NULL;
 	GtkTreeModel *model;
 	GPtrArray *query;
+	gboolean is_all_query, sorting_matches;
+	const char *current_sorting_type;
+
+	rhythmdb_read_lock (source->priv->db);
 
 	if (source->priv->active_query) {
 		g_signal_handlers_disconnect_by_func (G_OBJECT (source->priv->active_query),
-						      G_CALLBACK (query_complete_cb), source);
-		if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_GENRE)
-			g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-						  source->priv->genre_add_handler_id);
-		if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ARTIST)
-			g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-						  source->priv->artist_add_handler_id);
-		if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ALBUM)
-			g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-						  source->priv->album_add_handler_id);
+						      G_CALLBACK (query_complete_cb),
+						      source);
+		rhythmdb_model_cancel (RHYTHMDB_MODEL (source->priv->active_query));
 	}
-		
+
+	is_all_query = (source->priv->selected_genres == NULL &&
+			source->priv->selected_artists == NULL &&	    
+			source->priv->selected_albums == NULL &&	    
+			source->priv->search_text == NULL);
+	current_sorting_type = rb_entry_view_get_sorting_type (source->priv->songs);
+	sorting_matches = source->priv->cached_sorting_type
+		&& !strcmp (source->priv->cached_sorting_type, current_sorting_type);
+	if (is_all_query) {
+		if (sorting_matches) {
+			rb_debug ("cached query hit");
+			source->priv->model = source->priv->cached_all_query;
+			source->priv->active_query = NULL;
+			rb_entry_view_set_model (source->priv->songs,
+						 RHYTHMDB_MODEL (source->priv->cached_all_query));
+			g_object_set (G_OBJECT (source->priv->cached_genres_model),
+				      "query-model", source->priv->cached_all_query, NULL);
+			g_object_set (G_OBJECT (source->priv->cached_artists_model),
+				      "query-model", source->priv->cached_all_query, NULL);
+			g_object_set (G_OBJECT (source->priv->cached_albums_model),
+				      "query-model", source->priv->cached_all_query, NULL);
+			rb_property_view_set_model (source->priv->genres,
+						    source->priv->cached_genres_model);
+			rb_property_view_set_model (source->priv->artists,
+						    source->priv->cached_artists_model);
+			rb_property_view_set_model (source->priv->albums,
+						    source->priv->cached_albums_model);
+			rb_entry_view_poll_model (source->priv->songs);
+			rhythmdb_read_unlock (source->priv->db);
+			return;
+		} else if (source->priv->cached_all_query) {
+			rb_debug ("sorting mismatch, freeing cached query");
+			g_object_unref (source->priv->cached_all_query);
+			g_object_unref (source->priv->cached_genres_model);
+			g_object_unref (source->priv->cached_artists_model);
+			g_object_unref (source->priv->cached_albums_model);
+			g_free (source->priv->cached_sorting_type);
+		}
+	}
+
 	source->priv->query_type = qtype;
 
-	g_signal_handler_block (G_OBJECT (source->priv->songs),
-				source->priv->genre_add_handler_id);
-	g_signal_handler_block (G_OBJECT (source->priv->songs),
-				source->priv->artist_add_handler_id);
-	g_signal_handler_block (G_OBJECT (source->priv->songs),
-				source->priv->album_add_handler_id);
+	if (source->priv->cached_all_query == NULL
+	    || (is_all_query && !sorting_matches)) {
+		rb_debug ("caching new query");
+		query_model = source->priv->active_query = source->priv->model = 
+			source->priv->cached_all_query = rhythmdb_query_model_new_empty (source->priv->db);
+		source->priv->cached_sorting_type = g_strdup (current_sorting_type);
+		source->priv->cached_genres_model = rhythmdb_property_model_new (source->priv->db, RHYTHMDB_PROP_GENRE);
+		source->priv->cached_artists_model = rhythmdb_property_model_new (source->priv->db, RHYTHMDB_PROP_ARTIST);
+		source->priv->cached_albums_model = rhythmdb_property_model_new (source->priv->db, RHYTHMDB_PROP_ALBUM);
+		g_object_set (G_OBJECT (source->priv->cached_genres_model),
+			      "query-model", source->priv->cached_all_query, NULL);
+		g_object_set (G_OBJECT (source->priv->cached_artists_model),
+			      "query-model", source->priv->cached_all_query, NULL);
+		g_object_set (G_OBJECT (source->priv->cached_albums_model),
+			      "query-model", source->priv->cached_all_query, NULL);
+		rb_property_view_set_model (source->priv->genres,
+					    source->priv->cached_genres_model);
+		rb_property_view_set_model (source->priv->artists,
+					    source->priv->cached_artists_model);
+		rb_property_view_set_model (source->priv->albums,
+					    source->priv->cached_albums_model);
+		g_object_ref (G_OBJECT (source->priv->cached_all_query));
+		g_object_ref (G_OBJECT (source->priv->cached_genres_model));
+		g_object_ref (G_OBJECT (source->priv->cached_artists_model));
+		g_object_ref (G_OBJECT (source->priv->cached_albums_model));
+	} else {
+		if (qtype < RB_LIBRARY_QUERY_TYPE_GENRE) {
+			rb_property_view_reset (source->priv->genres);
+			g_list_foreach (source->priv->selected_genres, (GFunc) g_free, NULL);
+			g_list_free (source->priv->selected_genres);
+			source->priv->selected_genres = NULL;
+		}
+		genre_model = rb_property_view_get_model (source->priv->genres);
+		if (qtype < RB_LIBRARY_QUERY_TYPE_ARTIST) {
+			rb_property_view_reset (source->priv->artists);
+			g_list_foreach (source->priv->selected_artists, (GFunc) g_free, NULL);
+			g_list_free (source->priv->selected_artists);
+			source->priv->selected_artists = NULL;
+		}
+		artist_model = rb_property_view_get_model (source->priv->artists);
+		if (qtype < RB_LIBRARY_QUERY_TYPE_ALBUM) {
+			rb_property_view_reset (source->priv->albums);
+			g_list_foreach (source->priv->selected_albums, (GFunc) g_free, NULL);
+			g_list_free (source->priv->selected_albums);
+			source->priv->selected_albums = NULL;
+		}
+		album_model = rb_property_view_get_model (source->priv->albums);
 
-	if (qtype < RB_LIBRARY_QUERY_TYPE_GENRE) {
-		rb_property_view_reset (source->priv->genres);
-		g_list_foreach (source->priv->selected_genres, (GFunc) g_free, NULL);
-		g_list_free (source->priv->selected_genres);
-		source->priv->selected_genres = NULL;
-		g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-					  source->priv->genre_add_handler_id);
-	}
-	if (qtype < RB_LIBRARY_QUERY_TYPE_ARTIST) {
-		rb_property_view_reset (source->priv->artists);
-		g_list_foreach (source->priv->selected_artists, (GFunc) g_free, NULL);
-		g_list_free (source->priv->selected_artists);
-		source->priv->selected_artists = NULL;
-		g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-					  source->priv->artist_add_handler_id);
-	}
-	if (qtype < RB_LIBRARY_QUERY_TYPE_ALBUM) {
-		rb_property_view_reset (source->priv->albums);
-		g_list_foreach (source->priv->selected_albums, (GFunc) g_free, NULL);
-		g_list_free (source->priv->selected_albums);
-		source->priv->selected_albums = NULL;
-		g_signal_handler_unblock (G_OBJECT (source->priv->songs),
-					  source->priv->album_add_handler_id);
-	}
+		source->priv->active_query = source->priv->model = 
+			query_model = rhythmdb_query_model_new_empty (source->priv->db);
+		if (qtype < RB_LIBRARY_QUERY_TYPE_GENRE)
+			g_object_set (G_OBJECT (genre_model), "query-model", query_model, NULL);
+		else
+			g_object_set (G_OBJECT (genre_model), "query-model", NULL, NULL);
+	
+		if (qtype < RB_LIBRARY_QUERY_TYPE_ARTIST)
+			g_object_set (G_OBJECT (artist_model), "query-model", query_model, NULL);
+		else
+			g_object_set (G_OBJECT (artist_model), "query-model", NULL, NULL);
 
-	source->priv->active_query = source->priv->model = 
-		query_model = rhythmdb_query_model_new_empty (source->priv->db);
+		if (qtype < RB_LIBRARY_QUERY_TYPE_ALBUM)
+			g_object_set (G_OBJECT (album_model), "query-model", query_model, NULL);
+		else
+			g_object_set (G_OBJECT (album_model), "query-model", NULL, NULL);
+	}
 
 	model = GTK_TREE_MODEL (query_model);
 
-	g_signal_connect (G_OBJECT (query_model),
-			  "complete", G_CALLBACK (query_complete_cb),
-			  source);
+	g_signal_connect_object (G_OBJECT (query_model),
+				 "complete", G_CALLBACK (query_complete_cb),
+				 source, 0);
 	
 	rb_entry_view_set_model (source->priv->songs, RHYTHMDB_MODEL (query_model));
 
+	rhythmdb_read_unlock (source->priv->db);
+
 	query = construct_query_from_selection (source);
 	
 	if (!sync)
 		rhythmdb_do_full_query_async_parsed (source->priv->db, model, query);
-	else {
+	else
 		rhythmdb_do_full_query_parsed (source->priv->db, model, query);
-	}
-
+		
 	rhythmdb_query_free (query);
 	g_object_unref (G_OBJECT (query_model));
 
 	rb_entry_view_poll_model (source->priv->songs);
 }
 
-struct RBLibrarySourceQueryCompleteData
-{
-	RhythmDBQueryModel *model;
-	RBLibrarySource *source;
-};
-
-static gboolean
-idle_do_query_complete (struct RBLibrarySourceQueryCompleteData *data)
-{
-	GDK_THREADS_ENTER ();
-	if (data->source->priv->active_query != data->model)
-		goto out;
-
-	if (data->source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_GENRE)
-		g_signal_handler_unblock (G_OBJECT (data->source->priv->songs),
-					  data->source->priv->genre_add_handler_id);
-	if (data->source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ARTIST)
-		g_signal_handler_unblock (G_OBJECT (data->source->priv->songs),
-					  data->source->priv->artist_add_handler_id);
-	if (data->source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ALBUM)
-		g_signal_handler_unblock (G_OBJECT (data->source->priv->songs),
-					  data->source->priv->album_add_handler_id);
-	data->source->priv->active_query = NULL;
-out:
-	g_object_unref (G_OBJECT (data->model));
-	g_free (data);
-	GDK_THREADS_LEAVE ();
-	return FALSE;
-}
-
 static void
 query_complete_cb (RhythmDBQueryModel *model, RBLibrarySource *source)
 {
-	rb_debug ("query complete");
-	struct RBLibrarySourceQueryCompleteData *data
-		= g_new0 (struct RBLibrarySourceQueryCompleteData, 1);
-	data->model = model;
-	data->source = source;
-	g_object_ref (G_OBJECT (model));
-	g_idle_add ((GSourceFunc) idle_do_query_complete, data);
+	RhythmDBPropertyModel *genre_model = rb_property_view_get_model (source->priv->genres);
+	RhythmDBPropertyModel *artist_model = rb_property_view_get_model (source->priv->artists);
+	RhythmDBPropertyModel *album_model = rb_property_view_get_model (source->priv->albums);
+
+	rb_debug ("query complete; resetting data models");
+
+	if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_GENRE)
+		g_object_set (G_OBJECT (genre_model), "query-model", model, NULL);
+	if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ARTIST)
+		g_object_set (G_OBJECT (artist_model), "query-model", model, NULL);
+	if (source->priv->query_type >= RB_LIBRARY_QUERY_TYPE_ALBUM)
+		g_object_set (G_OBJECT (album_model), "query-model", model, NULL);
+	if (genre_model != source->priv->cached_genres_model)
+		g_object_set (G_OBJECT (source->priv->cached_genres_model),
+			      "query-model", model, NULL);
+	if (artist_model != source->priv->cached_artists_model)
+		g_object_set (G_OBJECT (source->priv->cached_artists_model),
+			      "query-model", model, NULL);
+	if (album_model != source->priv->cached_albums_model)
+		g_object_set (G_OBJECT (source->priv->cached_albums_model),
+			      "query-model", model, NULL);
+	source->priv->active_query = NULL;
 }
 
 /* static void */
