@@ -2,7 +2,7 @@
  *  arch-tag: Implementation of playlist source object
  *
  *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
- *  Copyright (C) 2003 Colin Walters <cwalters@gnome.org>
+ *  Copyright (C) 2003 Colin Walters <walters@gnome.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -67,6 +67,7 @@ static const char *impl_get_status (RBSource *source);
 static const char *impl_get_browser_key (RBSource *source);
 static GdkPixbuf *impl_get_pixbuf (RBSource *source);
 static RBEntryView *impl_get_entry_view (RBSource *source);
+static gboolean impl_can_cut (RBSource *asource);
 static GList * impl_cut (RBSource *source);
 static void impl_paste (RBSource *asource, GList *entries);
 static void impl_delete (RBSource *source);
@@ -170,9 +171,9 @@ rb_playlist_source_class_init (RBPlaylistSourceClass *klass)
 	source_class->impl_get_entry_view = impl_get_entry_view;
 	source_class->impl_can_rename = (RBSourceFeatureFunc) rb_true_function;
 	source_class->impl_can_search = (RBSourceFeatureFunc) rb_false_function;
-	source_class->impl_can_cut = (RBSourceFeatureFunc) rb_true_function;
+	source_class->impl_can_cut = impl_can_cut;
 	source_class->impl_can_copy = (RBSourceFeatureFunc) rb_true_function;
-	source_class->impl_can_delete = (RBSourceFeatureFunc) rb_true_function;
+	source_class->impl_can_delete = impl_can_cut;
 	source_class->impl_cut = impl_cut;
 	source_class->impl_paste = impl_paste;
 	source_class->impl_delete = impl_delete;
@@ -400,8 +401,13 @@ static void
 rb_playlist_source_entry_added_cb (RhythmDB *db, RhythmDBEntry *entry,
 				   RBPlaylistSource *source)
 {
-	const char *location = rhythmdb_entry_get_string (db, entry,
-							  RHYTHMDB_PROP_LOCATION);
+	const char *location;
+
+	if (source->priv->automatic)
+		return;
+	
+	location = rhythmdb_entry_get_string (db, entry,
+					      RHYTHMDB_PROP_LOCATION);
 	if (g_hash_table_lookup (source->priv->entries, location)) {
 		rhythmdb_query_model_add_entry (source->priv->model, entry);
 	}
@@ -414,8 +420,10 @@ rb_playlist_source_get_model (RBPlaylistSource *source)
 }
 
 void
-rb_playlist_source_set_query (RBPlaylistSource *source, GPtrArray *query,
-			      guint limit)
+rb_playlist_source_set_query (RBPlaylistSource *source,
+			      GPtrArray *query,
+			      guint limit_count,
+			      guint limit_mb)
 {
 	RhythmDBQueryModel *query_model;
 	GtkTreeModel *model;
@@ -424,7 +432,9 @@ rb_playlist_source_set_query (RBPlaylistSource *source, GPtrArray *query,
 
 	source->priv->model = query_model
 		= rhythmdb_query_model_new_empty (source->priv->db);
-	g_object_set (G_OBJECT (source->priv->model), "max-size", limit, NULL);
+	g_object_set (G_OBJECT (source->priv->model),
+		      "max-count", limit_count,
+		      "max-size", limit_mb, NULL);
 
 	model = GTK_TREE_MODEL (query_model);
 
@@ -470,36 +480,44 @@ impl_get_entry_view (RBSource *asource)
 	return source->priv->songs;
 }
 
+static gboolean
+impl_can_cut (RBSource *asource)
+{
+	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource);
+	return !source->priv->automatic;
+}
+
 static GList *
 impl_cut (RBSource *asource)
 {
-/* 	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource); */
-/* 	GList *sel = rb_node_view_get */
+	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource);
+	GList *sel = rb_entry_view_get_selected_entries (source->priv->songs);
+	GList *tem;
 
-/* 	for (; sel != NULL; sel = g_list_next (sel)) */
-/* 		rb_node_remove_child (source->priv->root, sel->data); */
+	for (tem = sel; tem; tem = tem->next)
+		rb_playlist_source_remove_entry (source, tem->data);
 
-/* 	return sel; */
-	return NULL;
+	return sel;
 }
 
 static void
-impl_paste (RBSource *asource, GList *nodes)
+impl_paste (RBSource *asource, GList *entries)
 {
-/* 	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource); */
+	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource);
 
-/* 	for (; nodes; nodes = g_list_next (nodes)) */
-/* 		rb_playlist_source_add_node (source, nodes->data); */
+	for (; entries; entries = g_list_next (entries))
+		rb_playlist_source_add_entry (source, entries->data);
 }
 
 static void
 impl_delete (RBSource *asource)
 {
 	RBPlaylistSource *source = RB_PLAYLIST_SOURCE (asource);
-	GList *l;
-
-	for (l = rb_entry_view_get_selected_entries (source->priv->songs); l != NULL; l = g_list_next (l))
-		rb_playlist_source_remove_entry (source, l->data);
+	GList *sel, *tem;
+	sel = rb_entry_view_get_selected_entries (source->priv->songs);
+	for (tem = sel; tem != NULL; tem = tem->next)
+		rb_playlist_source_remove_entry (source, tem->data);
+	g_list_free (sel);
 }
 
 static void
@@ -571,16 +589,20 @@ rb_playlist_source_add_location (RBPlaylistSource *source,
 {
 	RhythmDBEntry *entry;
 
-	if (g_hash_table_lookup (source->priv->entries, location))
+	rhythmdb_read_lock (source->priv->db);
+	if (g_hash_table_lookup (source->priv->entries, location)) {
+		rhythmdb_read_unlock (source->priv->db);
 		return;
+	}
 
 	g_hash_table_insert (source->priv->entries,
 			     g_strdup (location), GINT_TO_POINTER (1));
-	rhythmdb_read_lock (source->priv->db);
+
 	entry = rhythmdb_entry_lookup_by_location (source->priv->db, location);
 	if (entry)
 		rhythmdb_entry_ref_unlocked (source->priv->db, entry);
 	rhythmdb_read_unlock (source->priv->db);
+
 	if (entry != NULL) {
 		rhythmdb_query_model_add_entry (source->priv->model, entry);
 		rhythmdb_entry_unref (source->priv->db, entry);
@@ -702,8 +724,6 @@ rb_playlist_source_new_from_xml	(RhythmDB *db,
 	xmlNodePtr child;
 	char *tmp;
 
-	rhythmdb_read_lock (db);
-
 	source = RB_PLAYLIST_SOURCE (rb_playlist_source_new (db, FALSE));
 
 	tmp = xmlGetProp (node, "type");
@@ -714,21 +734,47 @@ rb_playlist_source_new_from_xml	(RhythmDB *db,
 	g_object_set (G_OBJECT (source), "name", tmp, NULL);
 	g_free (tmp);
 
-	for (child = node->children; child; child = child->next) {
-		char *location;
+	if (source->priv->automatic) {
+		GPtrArray *query;
+		gchar *limit_str;
+		guint limit_count = 0;
+		guint limit_mb = 0;
 
-		if (xmlNodeIsText (child))
-			continue;
+		child = node->children;
+		while (xmlNodeIsText (child))
+			child = child->next;
+
+		query = rhythmdb_query_deserialize (db, child);
+		limit_str = xmlGetProp (node, "limit-count");
+		if (!limit_str) /* Backwards compatibility */
+			limit_str = xmlGetProp (node, "limit");
+		if (limit_str) {
+			limit_count = atoi (limit_str);
+			g_free (limit_str);
+		}
+		limit_str = xmlGetProp (node, "limit-size");
+		if (limit_str) {
+			limit_mb = atoi (limit_str);
+			g_free (limit_str);
+		}
+		rb_playlist_source_set_query (source, query,
+					      limit_count,
+					      limit_mb);
+	} else {
+		for (child = node->children; child; child = child->next) {
+			char *location;
+
+			if (xmlNodeIsText (child))
+				continue;
 		
-		if (strcmp (child->name, "location"))
-			continue;
+			if (strcmp (child->name, "location"))
+				continue;
 		
-		location = xmlNodeGetContent (child);
-		rb_playlist_source_add_location (source, location);
+			location = xmlNodeGetContent (child);
+			rb_playlist_source_add_location (source, location);
+		}
 	}
 
-	rhythmdb_read_unlock (db);
-
 	return RB_SOURCE (source);
 }
 
@@ -768,6 +814,22 @@ rb_playlist_source_save_to_xml (RBPlaylistSource *source, xmlNodePtr parent_node
 			g_free (encoded);
 		} while (gtk_tree_model_iter_next (GTK_TREE_MODEL (source->priv->model),
 						   &iter));
-	} else
-		g_assert_not_reached ();
+	} else {
+		GPtrArray *query;
+		guint max_count;
+		guint max_size_mb;
+		char *limit_str;
+
+		g_object_get (G_OBJECT (source->priv->model),
+			      "max-count", &max_count,
+			      "max-size", &max_size_mb,
+			      "query", &query, NULL);
+		limit_str = g_strdup_printf ("%d", max_count);
+		xmlSetProp (node, "limit-count", limit_str);
+		g_free (limit_str);
+		limit_str = g_strdup_printf ("%d", max_size_mb);
+		xmlSetProp (node, "limit-size", limit_str);
+		g_free (limit_str);
+		rhythmdb_query_serialize (source->priv->db, query, node);
+	}
 }
