@@ -32,7 +32,7 @@
 #include <string.h>
 #include <stdlib.h>
 
-#include "eggtreemultidnd.h"
+#include "rb-tree-dnd.h"
 #include "rb-tree-view-column.h"
 #include "rb-entry-view.h"
 #include "rb-dialog.h"
@@ -87,6 +87,7 @@ static void rb_entry_view_row_changed_cb (GtkTreeModel *model,
 					  GtkTreeIter *iter,
 					  RBEntryView *view);
 static gboolean emit_entry_changed (RBEntryView *view);
+static void queue_changed_sig (RBEntryView *view);
 static void rb_entry_view_sync_columns_visible (RBEntryView *view);
 static void rb_entry_view_columns_config_changed_cb (GConfClient* client,
 						    guint cnxn_id,
@@ -133,6 +134,7 @@ struct RBEntryViewPrivate
 	gboolean playing;
 	RhythmDBModel *playing_model;
 	RhythmDBEntry *playing_entry;
+	gboolean playing_entry_in_view;
 	GtkTreeIter playing_entry_iter;
 
 	gboolean is_drag_source;
@@ -149,15 +151,11 @@ struct RBEntryViewPrivate
 	struct RBEntryViewReverseSortingData *reverse_sorting_data;
 
 	gboolean have_selection;
-	GList *entry_selection;
 
 	gboolean keep_selection;
 
 	RhythmDBEntry *selected_entry;
 
-	glong total_duration;
-	GnomeVFSFileSize total_size;
-
 	gboolean change_sig_queued;
 	guint change_sig_id;
 
@@ -419,8 +417,6 @@ rb_entry_view_finalize (GObject *object)
 
 	g_source_remove (view->priv->model_poll_id);
 
-	g_list_free (view->priv->entry_selection);
-
 	if (view->priv->gconf_notification_id > 0)
 		eel_gconf_notification_remove (view->priv->gconf_notification_id);
 	if (view->priv->sorting_gconf_notification_id > 0)
@@ -535,15 +531,18 @@ rb_entry_view_set_property (GObject *object,
 		break;
 	case PROP_MODEL:
 	{
-		gboolean sortable;
-		GList *tem;
 		RhythmDBModel *new_model;
 		struct RBEntryViewColumnSortData *sort_data;
 		
-		if (view->priv->model)
+		if (view->priv->model) {
 			rhythmdb_model_cancel (view->priv->model);
+			rhythmdb_query_model_set_connected (RHYTHMDB_QUERY_MODEL (view->priv->model), FALSE);
+		}
 		new_model = g_value_get_object (value);
 
+		rhythmdb_query_model_set_connected (RHYTHMDB_QUERY_MODEL (new_model),
+						    TRUE);
+
 		g_signal_connect_object (G_OBJECT (new_model),
 					 "row_inserted",
 					 G_CALLBACK (rb_entry_view_row_inserted_cb),
@@ -560,11 +559,7 @@ rb_entry_view_set_property (GObject *object,
 					 view,
 					 0);
 
-		sortable = rhythmdb_model_sortable (new_model);
-		for (tem = view->priv->clickable_columns; tem; tem = tem->next)
-			gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (tem->data), sortable);
-
-		if (sortable && view->priv->sorting_column) {
+		if (view->priv->sorting_column) {
 			sort_data = g_hash_table_lookup (view->priv->column_sort_data_map,
 							 view->priv->sorting_column);
 			g_assert (sort_data);
@@ -588,9 +583,7 @@ rb_entry_view_set_property (GObject *object,
 		gtk_tree_view_set_model (GTK_TREE_VIEW (view->priv->treeview),
 					 GTK_TREE_MODEL (new_model));
 		view->priv->model = new_model;
-
-		view->priv->total_duration = 0;
-		view->priv->total_size = 0;
+		queue_changed_sig (view);
 
 		break;
 	}
@@ -603,7 +596,8 @@ rb_entry_view_set_property (GObject *object,
 
 		entry = g_value_get_pointer (value);
 		
-		if (view->priv->playing_entry != NULL) {
+		if (view->priv->playing_entry != NULL
+		    && view->priv->playing_entry_in_view) {
 			path = gtk_tree_model_get_path (GTK_TREE_MODEL (view->priv->playing_model),
 							&view->priv->playing_entry_iter);
 			gtk_tree_model_row_changed (GTK_TREE_MODEL (view->priv->playing_model),
@@ -617,19 +611,23 @@ rb_entry_view_set_property (GObject *object,
 		view->priv->playing_model = view->priv->model;
 
 		if (view->priv->playing_entry != NULL) {
-			rhythmdb_model_entry_to_iter (view->priv->model,
-							view->priv->playing_entry,
-							&view->priv->playing_entry_iter);
-			path = gtk_tree_model_get_path (GTK_TREE_MODEL (view->priv->model),
-							&view->priv->playing_entry_iter);
-			gtk_tree_model_row_changed (GTK_TREE_MODEL (view->priv->model),
-						    path, &view->priv->playing_entry_iter);
-			gtk_tree_path_free (path);
+			view->priv->playing_entry_in_view = 
+				rhythmdb_model_entry_to_iter (view->priv->model,
+							  view->priv->playing_entry,
+							  &view->priv->playing_entry_iter);
+			if (view->priv->playing_entry_in_view) {
+				path = gtk_tree_model_get_path (GTK_TREE_MODEL (view->priv->model),
+								&view->priv->playing_entry_iter);
+				gtk_tree_model_row_changed (GTK_TREE_MODEL (view->priv->model),
+							    path, &view->priv->playing_entry_iter);
+				gtk_tree_path_free (path);
+			}
 		}
 
 		rb_entry_view_thaw (view);
 
 		if (view->priv->playing_entry
+		    && view->priv->playing_entry_in_view
 		    && !rb_entry_view_entry_is_visible (view, view->priv->playing_entry, &iter))
 			rb_entry_view_scroll_to_iter (view, &iter);
 	}
@@ -716,13 +714,13 @@ rb_entry_view_busy (RBEntryView *view)
 glong
 rb_entry_view_get_duration (RBEntryView *view)
 {
-	return view->priv->total_duration;
+	return rhythmdb_query_model_get_duration (RHYTHMDB_QUERY_MODEL (view->priv->model));
 }
 
 GnomeVFSFileSize
 rb_entry_view_get_total_size (RBEntryView *view)
 {
-	return view->priv->total_size;
+	return rhythmdb_query_model_get_size (RHYTHMDB_QUERY_MODEL (view->priv->model));
 }
 
 static RhythmDBEntry *
@@ -774,8 +772,6 @@ rb_entry_view_artist_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	const char *b_str = NULL;
 	gint ret;
 
-	rhythmdb_read_lock (view->priv->db);
-
 	a_str = rhythmdb_entry_get_string (view->priv->db, a, RHYTHMDB_PROP_ARTIST_SORT_KEY);
 	b_str = rhythmdb_entry_get_string (view->priv->db, b, RHYTHMDB_PROP_ARTIST_SORT_KEY);
 
@@ -808,22 +804,18 @@ rb_entry_view_artist_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	ret = 0;
 
 out:
-	rhythmdb_read_unlock (view->priv->db);
-
 	return ret;
 }
 
 static gint
-rb_entry_view_track_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
+rb_entry_view_album_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 			       RBEntryView *view)
 {
 	gint a_int, b_int;
 	const char *a_str = NULL;
-	const char *b_str = NULL;	
+	const char *b_str = NULL;
 	gint ret;
 
-	rhythmdb_read_lock (view->priv->db);
-
 	a_str = rhythmdb_entry_get_string (view->priv->db, a, RHYTHMDB_PROP_ALBUM_SORT_KEY);
 	b_str = rhythmdb_entry_get_string (view->priv->db, b, RHYTHMDB_PROP_ALBUM_SORT_KEY);
 
@@ -840,8 +832,34 @@ rb_entry_view_track_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	}
 
 out:
-	rhythmdb_read_unlock (view->priv->db);
+	return ret;
+}
 
+static gint
+rb_entry_view_track_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
+			       RBEntryView *view)
+{
+	gint a_int, b_int;
+	const char *a_str = NULL;
+	const char *b_str = NULL;	
+	gint ret;
+
+	a_str = rhythmdb_entry_get_string (view->priv->db, a, RHYTHMDB_PROP_ALBUM_SORT_KEY);
+	b_str = rhythmdb_entry_get_string (view->priv->db, b, RHYTHMDB_PROP_ALBUM_SORT_KEY);
+
+	ret = strcmp (a_str, b_str);
+	if (ret != 0)
+		goto out;
+
+	a_int = rhythmdb_entry_get_int (view->priv->db, a, RHYTHMDB_PROP_TRACK_NUMBER);
+	b_int = rhythmdb_entry_get_int (view->priv->db, b, RHYTHMDB_PROP_TRACK_NUMBER);
+
+	if (a_int != b_int) {
+		ret = (a_int < b_int ? -1 : 1);
+		goto out;
+	}
+
+out:
 	return ret;
 }
 
@@ -853,15 +871,11 @@ rb_entry_view_int_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	gint a_val, b_val;
 	gint ret;
 
-	rhythmdb_read_lock (data->view->priv->db);
-
 	a_val = rhythmdb_entry_get_int (data->view->priv->db, a, data->propid);
 	b_val = rhythmdb_entry_get_int (data->view->priv->db, b, data->propid);
 
 	ret = (a_val == b_val ? 0 : (a_val > b_val ? 1 : -1));
 
-	rhythmdb_read_unlock (data->view->priv->db);
-
 	return ret;
 }
 
@@ -872,15 +886,11 @@ rb_entry_view_long_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	glong a_val, b_val;
 	gint ret;
 
-	rhythmdb_read_lock (data->view->priv->db);
-
 	a_val = rhythmdb_entry_get_long (data->view->priv->db, a, data->propid);
 	b_val = rhythmdb_entry_get_long (data->view->priv->db, b, data->propid);
 
 	ret = (a_val == b_val ? 0 : (a_val > b_val ? 1 : -1));
 
-	rhythmdb_read_unlock (data->view->priv->db);
-
 	return ret;
 }
 
@@ -892,15 +902,11 @@ rb_entry_view_string_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	const char *b_val;	
 	gint ret;
 
-	rhythmdb_read_lock (data->view->priv->db);
-
 	a_val = rhythmdb_entry_get_string (data->view->priv->db, a, data->propid);
 	b_val = rhythmdb_entry_get_string (data->view->priv->db, b, data->propid);
 
 	ret = strcmp (a_val, b_val);
 
-	rhythmdb_read_unlock (data->view->priv->db);
-
 	return ret;
 }
 
@@ -1101,6 +1107,12 @@ free_out:
 	g_strfreev (strs);
 }
 
+const char *
+rb_entry_view_get_sorting_type (RBEntryView *view)
+{
+	return eel_gconf_get_string (view->priv->sorting_key);
+}
+
 static void
 rb_entry_view_column_clicked_cb (GtkTreeViewColumn *column, RBEntryView *view)
 {
@@ -1165,6 +1177,7 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 		sort_data = g_new0 (struct RBEntryViewCellDataFuncData, 1);
 		sort_data->view = view;
 		sort_data->propid = propid;
+		sort_func = (GCompareDataFunc) rb_entry_view_int_sort_func;
 
 		renderer = rb_cell_renderer_rating_new ();
 		gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -1174,6 +1187,7 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 							 view,
 							 NULL);
 		gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
+		gtk_tree_view_column_set_clickable (column, TRUE);
 		gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &width, NULL);
 		gtk_tree_view_column_set_fixed_width (column, width * 5 + 5);
 		g_signal_connect_object (renderer,
@@ -1229,7 +1243,8 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 		cell_data->propid = propid;
 		cell_data_func = (GtkTreeCellDataFunc) rb_entry_view_string_cell_data_func;				
 		sort_data->propid = RHYTHMDB_PROP_ALBUM_SORT_KEY;
-		sort_func = (GCompareDataFunc) rb_entry_view_string_sort_func;
+		sort_func = (GCompareDataFunc) rb_entry_view_album_sort_func;
+		real_sort_data = view;
 		title = _("A_lbum");
 		key = "Album";
 		rb_tree_view_column_set_expand (RB_TREE_VIEW_COLUMN (column), TRUE);
@@ -1302,7 +1317,7 @@ rb_entry_view_append_column_custom (RBEntryView *view,
 	gtk_tree_view_column_set_title (column, title);
 	gtk_tree_view_column_set_reorderable (column, FALSE);
 
-	if (gtk_tree_view_column_get_clickable)
+	if (gtk_tree_view_column_get_clickable (column))
 		view->priv->clickable_columns = g_list_append (view->priv->clickable_columns, column);
 
 	g_signal_connect_object (G_OBJECT (column), "clicked",
@@ -1342,7 +1357,7 @@ rb_entry_view_constructor (GType type, guint n_construct_properties,
 {
 	RBEntryView *view;
 	RBEntryViewClass *klass;
-	GObjectClass *parent_class;  
+	GObjectClass *parent_class;
 	klass = RB_ENTRY_VIEW_CLASS (g_type_class_peek (type));
 
 	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
@@ -1372,26 +1387,29 @@ rb_entry_view_constructor (GType type, guint n_construct_properties,
 	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (view->priv->treeview), TRUE);
 	gtk_tree_view_set_reorderable (GTK_TREE_VIEW (view->priv->treeview), TRUE);
 	gtk_tree_selection_set_mode (view->priv->selection, GTK_SELECTION_MULTIPLE);
-	egg_tree_multi_drag_add_drag_support (GTK_TREE_VIEW (view->priv->treeview));
-
-	if (view->priv->is_drag_source)
-		gtk_tree_view_enable_model_drag_source (GTK_TREE_VIEW (view->priv->treeview),
-							GDK_BUTTON1_MASK,
-							rb_entry_view_drag_types,
-							G_N_ELEMENTS (rb_entry_view_drag_types),
-							GDK_ACTION_COPY);
-	if (view->priv->is_drag_dest)
-		gtk_tree_view_enable_model_drag_dest (GTK_TREE_VIEW (view->priv->treeview),
-						      rb_entry_view_drag_types,
-						      G_N_ELEMENTS (rb_entry_view_drag_types),
-						      GDK_ACTION_COPY);
 	
+	if (view->priv->is_drag_source)
+		rb_tree_dnd_add_drag_source_support (GTK_TREE_VIEW (view->priv->treeview),
+						     GDK_BUTTON1_MASK,
+						     rb_entry_view_drag_types,
+						     G_N_ELEMENTS (rb_entry_view_drag_types),
+						     GDK_ACTION_COPY | GDK_ACTION_MOVE);
+	if (view->priv->is_drag_dest)
+		rb_tree_dnd_add_drag_dest_support (GTK_TREE_VIEW (view->priv->treeview),
+						   RB_TREE_DEST_CAN_DROP_BETWEEN,
+						   rb_entry_view_drag_types,
+						   G_N_ELEMENTS (rb_entry_view_drag_types),
+						   GDK_ACTION_COPY | GDK_ACTION_MOVE);
+
 	gtk_container_add (GTK_CONTAINER (view), view->priv->treeview);
 
 	{
 		GtkTreeViewColumn *column;
+		GtkTooltips *tooltip;
 		GtkCellRenderer *renderer;
 		guint width;
+
+		tooltip = gtk_tooltips_new ();
 		
 		/* Playing icon column */
 		column = GTK_TREE_VIEW_COLUMN (rb_tree_view_column_new ());
@@ -1406,6 +1424,9 @@ rb_entry_view_constructor (GType type, guint n_construct_properties,
 		gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &width, NULL);
 		gtk_tree_view_column_set_fixed_width (column, width + 5);
 		gtk_tree_view_append_column (GTK_TREE_VIEW (view->priv->treeview), column);
+
+		gtk_tooltips_set_tip (GTK_TOOLTIPS (tooltip), GTK_WIDGET (column->button),
+				       _("Now Playing"), NULL);
 	}
 
 	view->priv->gconf_notification_id = 
@@ -1462,16 +1483,12 @@ rb_entry_view_rated_cb (RBCellRendererRating *cellrating,
 
 	rb_entry_view_freeze (view);
 
-	rhythmdb_write_lock (view->priv->db);
-
 	g_value_init (&value, G_TYPE_INT);
 	g_value_set_int (&value, rating);
-	rhythmdb_entry_set (view->priv->db, entry, RHYTHMDB_PROP_RATING,
-			    &value);
+	rhythmdb_entry_queue_set (view->priv->db, entry, RHYTHMDB_PROP_RATING,
+				  &value);
 	g_value_unset (&value);
 
-	rhythmdb_write_unlock (view->priv->db);
-
 	rb_entry_view_thaw (view);
 }
 
@@ -1589,8 +1606,6 @@ rb_entry_view_get_selected_entries (RBEntryView *view)
 	gtk_tree_selection_selected_foreach (view->priv->selection,
 					     (GtkTreeSelectionForeachFunc) harvest_entries,
 					     (gpointer) &list);
-	g_list_free (view->priv->entry_selection);
-	view->priv->entry_selection = g_list_copy (list);
 
 	return list;
 }
@@ -1716,23 +1731,6 @@ rb_entry_view_row_inserted_cb (GtkTreeModel *model,
 			       GtkTreeIter *iter,
 			       RBEntryView *view)
 {
-	RhythmDBEntry *entry = entry_from_tree_iter (view, iter);
-	glong duration;
-	glong size;
-	
-	rhythmdb_read_lock (view->priv->db);
-	
-	duration = rhythmdb_entry_get_long (view->priv->db, entry,
-					    RHYTHMDB_PROP_DURATION);
-	size = rhythmdb_entry_get_long (view->priv->db, entry,
-					RHYTHMDB_PROP_FILE_SIZE);
-
-	rhythmdb_read_unlock (view->priv->db);
-
-	view->priv->total_duration += duration;
-	view->priv->total_size += size;
-	
-	g_signal_emit (G_OBJECT (view), rb_entry_view_signals[ENTRY_ADDED], 0, entry);
 	queue_changed_sig (view);
 }
 
@@ -1742,8 +1740,6 @@ rb_entry_view_row_deleted_cb (GtkTreeModel *model,
 			      RBEntryView *view)
 {
 	RhythmDBEntry *entry = entry_from_tree_path (view, path);
-	glong duration;
-	glong size;
 
 	if (entry == view->priv->playing_entry) {
 		view->priv->playing_entry = NULL;
@@ -1754,18 +1750,6 @@ rb_entry_view_row_deleted_cb (GtkTreeModel *model,
 			       0, view->priv->playing_entry);
 	}
 	
-	rhythmdb_read_lock (view->priv->db);
-	
-	duration = rhythmdb_entry_get_long (view->priv->db, entry,
-					    RHYTHMDB_PROP_DURATION);
-	size = rhythmdb_entry_get_long (view->priv->db, entry,
-					RHYTHMDB_PROP_FILE_SIZE);
-
-	rhythmdb_read_unlock (view->priv->db);
-
-	view->priv->total_duration -= duration;
-	view->priv->total_size -= size;
-	
 	rb_debug ("row deleted");
 	g_signal_emit (G_OBJECT (view), rb_entry_view_signals[ENTRY_DELETED], 0, entry);
 	queue_changed_sig (view);
@@ -1819,10 +1803,10 @@ rb_entry_view_select_entry (RBEntryView *view,
 
 	rb_entry_view_select_none (view);
 
-	rhythmdb_model_entry_to_iter (view->priv->model,
-					entry, &iter);
-
-	gtk_tree_selection_select_iter (view->priv->selection, &iter);
+	if (rhythmdb_model_entry_to_iter (view->priv->model,
+					  entry, &iter)) {
+		gtk_tree_selection_select_iter (view->priv->selection, &iter);
+	}
 
 	view->priv->selection_lock = FALSE;
 }
@@ -1833,10 +1817,10 @@ rb_entry_view_scroll_to_entry (RBEntryView *view,
 {
 	GtkTreeIter iter;
 	
-	rhythmdb_model_entry_to_iter (view->priv->model,
-					entry, &iter);
-	
-	rb_entry_view_scroll_to_iter (view, &iter);
+	if (rhythmdb_model_entry_to_iter (view->priv->model,
+					  entry, &iter)) {
+		rb_entry_view_scroll_to_iter (view, &iter);
+	}
 }
 
 static void
@@ -1880,8 +1864,9 @@ rb_entry_view_entry_is_visible (RBEntryView *view,
 	if (!GTK_WIDGET_REALIZED (view))
 		return FALSE;
 
-	rhythmdb_model_entry_to_iter (view->priv->model,
-					entry, iter);
+	if (!rhythmdb_model_entry_to_iter (view->priv->model,
+					   entry, iter))
+		return FALSE;
 
 	path = gtk_tree_model_get_path (GTK_TREE_MODEL (view->priv->model), iter);
 	gtk_tree_view_get_cell_area (GTK_TREE_VIEW (view->priv->treeview),
@@ -1923,11 +1908,9 @@ rb_entry_view_enable_drag_source (RBEntryView *view,
 {
 	g_return_if_fail (view != NULL);
 
-	egg_tree_multi_drag_add_drag_support (GTK_TREE_VIEW (view->priv->treeview));
-
-	gtk_tree_view_enable_model_drag_source (GTK_TREE_VIEW (view->priv->treeview),
-						GDK_BUTTON1_MASK | GDK_BUTTON3_MASK,
-						targets, n_targets, GDK_ACTION_COPY);
+	rb_tree_dnd_add_drag_source_support (GTK_TREE_VIEW (view->priv->treeview),
+					 GDK_BUTTON1_MASK | GDK_BUTTON3_MASK,
+					 targets, n_targets, GDK_ACTION_COPY);
 }
 
 static void
